"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING, Callable
from hashlib import blake2b
from pathlib import Path

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.types import PATH_TYPE

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


# FILE_HASH_INCREMENTAL_THRESHOLD: int = 104857600  # 100mb
FILE_HASH_INCREMENTAL_THRESHOLD: int = 52428800  # 50mb


def file_hash(in_file: "PATH_TYPE", hash_algo: Callable = blake2b) -> str:
    in_file = Path(in_file)
    if not in_file.is_file():
        raise OSError(f"The path {in_file.as_posix()!r} either does not exist or is a Folder.")
    if in_file.stat().st_size > FILE_HASH_INCREMENTAL_THRESHOLD:
        _hash = hash_algo(usedforsecurity=False)
        with in_file.open("rb", buffering=FILE_HASH_INCREMENTAL_THRESHOLD // 4) as f:
            for chunk in f:
                _hash.update(chunk)
        return _hash.hexdigest()

    return hash_algo(in_file.read_bytes(), usedforsecurity=False).hexdigest()


# region[Main_Exec]

if __name__ == '__main__':
    print(file_hash(Path(r"D:\Dropbox\hobby\Modding\Programs\Github\My_Repos\Antistasi_Logbook\antistasi_logbook\storage\storage.db")))

# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import re
import logging
from typing import TYPE_CHECKING, Any, Union
from pathlib import Path
from logging.handlers import BaseRotatingHandler

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.enums import MiscEnum
from gidapptools.general_helper.conversion import human2bytes

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.types import PATH_TYPE

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


DEFAULT_MAX_BYTES = human2bytes("5 mb")


class _DefaultFileNameTemplate:

    def __init__(self, base_name: str) -> None:
        self.base_name = base_name
        self.backup_file_name_regex = re.compile(rf"{self.base_name}.log\_\d+", re.IGNORECASE)

    def format(self, **kwargs) -> str:
        return "{base_name}.log".format(base_name=self.base_name, **kwargs)

    def is_same_kind_log_file(self, other_file: Path) -> bool:
        return other_file.name.casefold() == "{base_name}.log".format(base_name=self.base_name).casefold()

    def is_backup_log_file(self, other_file: Path) -> bool:
        return self.backup_file_name_regex.match(other_file.name) is not None


class GidBaseRotatingFileHandler(BaseRotatingHandler):

    def __init__(self,
                 base_name: str,
                 log_folder: "PATH_TYPE",
                 file_name_template: Union[str, Any] = None,
                 backup_folder: "PATH_TYPE" = MiscEnum.AUTO,
                 rotate_on_start: bool = True,
                 backup_amount_limit: int = 3) -> None:
        self.base_name = base_name
        self.file_name_template = _DefaultFileNameTemplate(self.base_name) if file_name_template is None else file_name_template
        self.log_folder = Path(log_folder)
        self.backup_folder = self.log_folder.joinpath("old_logs") if backup_folder is MiscEnum.AUTO else Path(backup_folder)
        self.rotate_on_start = rotate_on_start
        self.backup_amount_limit = backup_amount_limit
        self.full_file_path: Path = self._construct_full_file_path()
        self.first_record_emited: bool = False
        super().__init__(self.full_file_path, "a", encoding="utf-8", delay=True, errors="ignore")

    def emit(self, record) -> None:
        with self.lock:
            if self.first_record_emited is False:
                self.on_start_rotation()
                self.first_record_emited = True

        return super().emit(record)

    def _construct_full_file_path(self) -> Path:
        name = self.file_name_template.format(number=0)
        full_path = self.log_folder.joinpath(name)

        return full_path

    def _get_old_logs(self) -> tuple[Path]:
        def _is_old_log(_in_file: Path) -> bool:
            return _in_file.is_file() and self.file_name_template.is_same_kind_log_file(_in_file)

        _out = tuple(file for file in tuple(self.log_folder.iterdir()) if _is_old_log(file) is True)

        return _out

    def _get_backup_logs(self) -> list[Path]:
        if self.backup_folder.exists() is False:
            return []

        def _is_old_backup(_in_file: Path) -> bool:
            return _in_file.is_file() and self.file_name_template.is_backup_log_file(_in_file)

        _out = sorted((file for file in self.backup_folder.iterdir() if _is_old_backup(file)), key=lambda x: x.stat().st_mtime)
        return _out

    def on_start_rotation(self) -> None:
        try:
            self.acquire()
            self.full_file_path.parent.mkdir(exist_ok=True, parents=True)
            for file in self._get_old_logs():
                self.move_file_to_backup_folder(file)
            self.remove_excess_backup_files()
        finally:
            self.release()

    def remove_excess_backup_files(self) -> None:
        if self.backup_amount_limit is None:
            return
        backup_logs = self._get_backup_logs()
        while len(backup_logs) > self.backup_amount_limit:
            to_delete: Path = backup_logs.pop(0)
            to_delete.unlink(missing_ok=True)

    def move_file_to_backup_folder(self, file: Path) -> None:
        self.backup_folder.mkdir(parents=True, exist_ok=True)
        number = 0
        target_path = self.backup_folder.joinpath(file.name)

        while target_path.exists() is True:
            number += 1
            name = f"{file.name}_{number}"
            target_path = self.backup_folder.joinpath(name)
        os.rename(src=file, dst=target_path)

    def shouldRollover(self, record: logging.LogRecord) -> bool:
        return False
# region[Main_Exec]


if __name__ == '__main__':
    pass
# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import json
from typing import Any
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.data.json import THIS_FILE_DIR as JSON_DATA_DIR

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()
WEBCOLORS_JSON = JSON_DATA_DIR.joinpath("webcolors.json")
# endregion[Constants]


def get_webcolors_data() -> list[dict[str:Any]]:
    with WEBCOLORS_JSON.open('r', encoding='utf-8', errors='ignore') as f:
        data = json.load(f)
    _out = []
    for item_data in data:
        new_item_data = {}
        new_item_data["name"] = item_data.get("name").casefold()
        new_item_data["r"] = item_data.get("rgb").get('r')
        new_item_data["g"] = item_data.get("rgb").get('g')
        new_item_data["b"] = item_data.get("rgb").get('b')
        new_item_data["alpha"] = 1.0
        _out.append(new_item_data)
    return _out
# region[Main_Exec]


if __name__ == '__main__':
    pass

# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import logging
from abc import ABC, abstractmethod
from pprint import pprint
from typing import TYPE_CHECKING, Union, Literal, Iterable
from pathlib import Path
from datetime import datetime, timezone, timedelta

# * Third Party Imports --------------------------------------------------------------------------------->
from tzlocal import get_localzone

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_logger.enums import LoggingLevel, LoggingSectionAlignment
from gidapptools.general_helper.string_helper import StringCase, StringCaseConverter

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gid_logger.records import LOG_RECORD_TYPES

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


def get_all_func_names(file: Path, recursive: True):
    import ast

    def _get_names(source_file: Path) -> list[str]:
        root = ast.parse(source_file.read_text(encoding='utf-8', errors='ignore'))
        return sorted({node.name for node in ast.walk(root) if isinstance(node, ast.FunctionDef)})

    def _process_item(_item: Path):
        if _item.is_file() and _item.suffix == '.py':
            yield _get_names(_item)
        elif _item.is_dir():
            for _sub_item in _item.iterdir():
                yield from _process_item(_sub_item)
    if recursive is False:
        names = _get_names(file)
    else:
        names = []
        for _names in _process_item(file.parent):
            names += _names
    return sorted(set(names), key=len)


def get_all_module_names(file: Path) -> list[str]:

    def _get_module_name(_item: Path) -> str:
        module_name = _item.stem
        return module_name

    def _process_item(_item: Path):
        if _item.is_file() and _item.suffix == '.py':
            yield _get_module_name(_item)
        elif _item.is_dir():
            for _sub_item in _item.iterdir():
                yield from _process_item(_sub_item)

    if file.is_file():
        file = file.parent
    return sorted({i for i in _process_item(file)}, key=len)


class AbstractLoggingStyleSection:
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.LEFT
    default_width: int = 0
    default_text: str = str(None)

    def __init__(self,
                 position: int = None,
                 width: int = None,
                 alignment: Union[LoggingSectionAlignment, str] = None) -> None:
        self.position = position
        self.width = width or self.default_width
        if alignment is None:
            self.alignment = self.default_alignment
        elif isinstance(alignment, str):
            self.alignment = LoggingSectionAlignment(alignment)
        else:
            self.alignment = alignment

    @abstractmethod
    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        ...

    def align(self, text: str) -> str:
        width = self.width() if callable(self.width) else self.width
        if len(text) > width:
            text = '...' + text[-(width - 3):]
        return self.alignment.align(text, width)

    def format(self, record: "LOG_RECORD_TYPES") -> str:
        try:
            text = str(self.get_formated_value(record=record))
        except AttributeError:
            text = self.default_text
        return self.align(text)

    @classmethod
    def ___section_name___(cls) -> str:
        return StringCaseConverter.convert_to(cls.__name__.removesuffix("Section"), StringCase.SNAKE)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(position={self.position!r})"


class TimeSection(AbstractLoggingStyleSection):

    local_timezone: timezone = get_localzone()
    default_time_format = '%Y-%m-%d %H:%M:%S'
    default_msec_format = '.{msec:03.0f}'
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.LEFT

    def __init__(self,
                 position: int = None,
                 time_zone: timezone = None,
                 time_format: str = None,
                 msec_format: str = None,
                 width: int = None,
                 alignment: Union[LoggingSectionAlignment, str] = None) -> None:
        super().__init__(position=position, alignment=alignment)

        self.time_zone = time_zone
        self.time_format = time_format or self.default_time_format
        self.msec_format = msec_format or self.default_msec_format
        self.width = width or (len(self.time_format) + len(self.msec_format.format(msec=0)) + 2)

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        time_value = datetime.fromtimestamp(record.created, tz=self.local_timezone)
        msec_value = record.msecs
        if msec_value == 1000:
            msec_value = 0
            time_value = time_value + timedelta(seconds=1)
        if self.time_zone is not None:
            time_value = time_value.replace(tzinfo=self.time_zone)
        if self.time_format.casefold() == "isoformat":
            return time_value.isoformat(timespec='seconds') + self.msec_format.format(msec=record.msecs)
        return time_value.strftime(self.time_format) + self.msec_format.format(msec=record.msecs)


class LevelSection(AbstractLoggingStyleSection):
    default_case: StringCase = StringCase.UPPER
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.CENTER
    default_width: int = max(len(i) for i in LoggingLevel._member_names_)

    def __init__(self,
                 position: int = None,
                 case: StringCase = None,
                 width: int = None,
                 alignment: Union[LoggingSectionAlignment, str] = None) -> None:
        super().__init__(position=position, alignment=alignment, width=width)
        if case is None:
            self.case = self.default_case
        elif isinstance(case, str):
            self.case = StringCase(case)
        else:
            self.case = case

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        return StringCaseConverter.convert_to(record.levelname, self.case)


class ThreadSection(AbstractLoggingStyleSection):
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.CENTER
    default_width: int = 20

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        return record.threadName


class LineNumberSection(AbstractLoggingStyleSection):
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.CENTER
    default_width: int = 5

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        return record.lineno


class PathSection(AbstractLoggingStyleSection):
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.LEFT

    def __init__(self, position: int = None,
                 depth: int = 2,
                 with_extension: bool = False,
                 width: int = None,
                 alignment: Union[LoggingSectionAlignment, str] = None) -> None:
        super().__init__(position=position, width=self.width_from_env, alignment=alignment)
        self.with_extension = with_extension
        self.depth = depth
        self._width_cache: int = None

    def width_from_env(self) -> int:
        if self._width_cache is None:
            self._width_cache = int(os.getenv("MAX_MODULE_NAME_LEN", "20"))
        return self._width_cache

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        path = Path(record.pathname)
        if self.with_extension is False:
            path = path.with_suffix('')
        return '.'.join(part for part in path.parts[-self.depth:])


class FunctionNameSection(AbstractLoggingStyleSection):
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.LEFT
    default_width: int = int(os.getenv("MAX_FUNC_NAME_LEN", "10"))
    default_text: str = "-"

    def __init__(self) -> None:
        super().__init__(width=self.width_from_env)
        self._width_cache: int = None

    def width_from_env(self) -> int:
        if self._width_cache is None:
            self._width_cache = int(os.getenv("MAX_FUNC_NAME_LEN", "10"))
        return self._width_cache

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        if record.funcName == "<module>":
            return self.default_text
        return record.funcName


class AbstractSectionLoggingStyle(ABC):
    default_separator: str = " | "
    default_message_start_indicator: str = ' ||--> '

    def __init__(self, sections: Iterable[Union[str, AbstractLoggingStyleSection]], separator: str = None, message_start_indicator: str = None, message_section: AbstractLoggingStyleSection = None):
        self.sections = sections
        self.sorted_sections: tuple[AbstractLoggingStyleSection] = self.sort_sections()
        self.separator = separator or self.default_separator
        self.message_start_indicator = message_start_indicator or self.default_message_start_indicator
        self.message_section = message_section

    def sort_sections(self) -> tuple[AbstractLoggingStyleSection]:
        temp_sorted = [i[0] for i in sorted([(v, idx) for idx, v in enumerate(self.sections)], key=lambda x: (x[0].position, x[1]))]
        for idx, sect in enumerate(temp_sorted):
            if sect.position is None:
                sect.position = idx
        return tuple(temp_sorted)

    def usesTime(self) -> bool:
        return False

    @abstractmethod
    def validate(self) -> None:
        ...

    @abstractmethod
    def _format(self, record: "LOG_RECORD_TYPES") -> str:
        ...

    def format(self, record: "LOG_RECORD_TYPES"):
        return self._format(record)


class GidSectionLoggingStyle(AbstractSectionLoggingStyle):

    def validate(self) -> None:
        if any(isinstance(i, AbstractLoggingStyleSection) is False for i in self.sections):
            raise ValueError("invalid format")

    def _format(self, record: "LOG_RECORD_TYPES") -> str:

        message_part = record.getMessage() if self.message_section is None else self.message_section.format(record)

        return self.separator.join(section.format(record) for section in self.sorted_sections) + self.message_start_indicator + message_part


class GidLoggingFormatter(logging.Formatter):
    default_fmt = (TimeSection(), LineNumberSection(), LevelSection(), ThreadSection(), PathSection(), FunctionNameSection())

    def __init__(self,
                 fmt: Union[str, Iterable[Union[str, AbstractLoggingStyleSection]]] = None,
                 style: Literal["section", "$", "{", "$"] = "section",
                 validate: bool = False,
                 **kwargs) -> None:
        if style == "section":
            if fmt is None:
                fmt = self.default_fmt
            self._style = GidSectionLoggingStyle(sections=fmt, separator=kwargs.pop("separator", None), message_start_indicator=kwargs.pop("message_start_indicator", None), message_section=kwargs.pop("message_section", None))
        else:
            self._style = logging._STYLES[style][0](fmt)
            if validate:
                self._style.validate()

            self._fmt = self._style._fmt
            self.datefmt = kwargs.pop("datefmt", None)

    def set_time(self, record: "LOG_RECORD_TYPES") -> "LOG_RECORD_TYPES":
        return record

    def format_message(self, record: "LOG_RECORD_TYPES") -> str:
        return self.formatMessage(record=record)

    def format_stack(self, stack_info):
        return self.formatStack(stack_info=stack_info)

    def format_exception(self, record: "LOG_RECORD_TYPES") -> "LOG_RECORD_TYPES":
        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(ei=record.exc_info)
        return record

    def format(self, record: "LOG_RECORD_TYPES") -> str:
        record = self.set_time(record=record)
        text = self.format_message(record=record)
        record = self.format_exception(record=record)
        if record.exc_text:
            if text[-1:] != "\n":
                text = text + "\n"
            text = text + record.exc_text
        if record.stack_info:
            if text[-1:] != "\n":
                text = text + "\n"
            text = text + self.format_stack(record.stack_info)
        return text

# region[Main_Exec]


if __name__ == '__main__':
    pprint(get_all_module_names(Path(r"D:\Dropbox\hobby\Modding\Programs\Github\My_Repos\GidAppTools\gidapptools\__main__.py")))
# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import shutil
from pprint import pformat
from typing import Any, Union, Callable, Optional
from pathlib import Path
from tempfile import mkdtemp
from contextlib import contextmanager

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.utility.enums import NamedMetaPath
from gidapptools.utility.helper import make_pretty
from gidapptools.utility._debug_tools import dprint
from gidapptools.abstract_classes.abstract_meta_item import AbstractMetaItem

print = dprint

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


class MetaPaths(AbstractMetaItem):

    def __init__(self, code_base_dir: Path, paths: dict[NamedMetaPath, Path]) -> None:
        self.code_base_dir = Path(code_base_dir).resolve()
        self._paths: dict[Union[NamedMetaPath, str], Optional[Path]] = paths
        self._created_normal_paths: set[Path] = set()
        self._created_temp_dirs: set[Path] = set()

    def get_path(self, identifier: Union[NamedMetaPath, str], default: Any = NotImplemented) -> Path:
        if isinstance(identifier, str):
            try:
                identifier = NamedMetaPath(identifier)
            except ValueError:
                pass
        path = self._paths.get(identifier, None)
        if path is None:
            return default

        if path.exists() is False:
            path.mkdir(parents=True, exist_ok=True)
            self._created_normal_paths.add(path)
        return path

    @property
    def data_dir(self) -> Path:
        return self.get_path(NamedMetaPath.DATA)

    @property
    def cache_dir(self) -> Path:
        return self.get_path(NamedMetaPath.CACHE)

    @property
    def temp_dir(self) -> Path:
        return self.get_path(NamedMetaPath.TEMP)

    @property
    def log_dir(self) -> Path:
        return self.get_path(NamedMetaPath.LOG)

    @property
    def config_dir(self) -> Path:
        return self.get_path(NamedMetaPath.CONFIG)

    @property
    def config_spec_dir(self) -> Path:
        return self.get_path(NamedMetaPath.CONFIG_SPEC)

    @property
    def db_dir(self) -> Path:
        return self.get_path(NamedMetaPath.DB)

    def get_new_temp_dir(self, suffix: str = None, name: str = None) -> Path:
        if name is not None:
            temp_dir = self.temp_dir.joinpath(name)
        else:
            temp_dir = Path(mkdtemp(dir=self.temp_dir, suffix=suffix))
        if temp_dir.exists() is False:
            temp_dir.mkdir(parents=True, exist_ok=True)
        self._created_temp_dirs.add(temp_dir)
        return temp_dir

    @contextmanager
    def context_new_temp_dir(self, suffix: str = None):
        temp_dir = self.get_new_temp_dir(suffix=suffix)
        yield temp_dir
        shutil.rmtree(temp_dir)

    def clean_all_temp(self) -> None:
        while len(self._created_temp_dirs) != 0:
            temp_dir = self._created_temp_dirs.pop()
            shutil.rmtree(temp_dir)

    def as_dict(self, pretty: bool = False) -> dict[str, Any]:

        _out = vars(self)
        if pretty is True:
            _out = make_pretty(self)

        return pformat(_out)

    def to_storager(self, storager: Callable = None) -> None:
        if storager is None:
            return
        storager(self)

    def clean_up(self, remove_all_paths: bool = False, **kwargs) -> None:
        self.clean_all_temp()
        if remove_all_paths is True:
            for path in self._created_normal_paths:
                if path.exists():
                    if kwargs.get('dry_run', False) is True:
                        print(f"Simulating deleting of path {path.as_posix()!r}, with 'shutil.rmtree'.")
                    else:
                        shutil.rmtree(path)
        # TODO: find a way to clean shit up, but completely, also add optional kwarg that also removes the author folder


# region[Main_Exec]
if __name__ == '__main__':
    pass
# endregion[Main_Exec]



# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import inspect
from time import time_ns, thread_time_ns, perf_counter_ns, process_time_ns
from typing import Union, Callable
from pathlib import Path
from functools import wraps
from threading import RLock
from contextlib import contextmanager

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.conversion import seconds2human

TIME_NS_FUNC_TYPE = Union[perf_counter_ns, process_time_ns, time_ns, thread_time_ns]


def _check_env_var_condition(second_var_name: str) -> bool:
    return any(os.getenv(env_name, '0').casefold() in {'1', "true"} for env_name in ["GIDAPPTOOLS_TIMING_ENABLED", "IS_DEV", second_var_name])


time_execution_path_locks: dict[Path, RLock] = {}


def get_time_execution_path_lock(path: Path) -> RLock:
    lock = time_execution_path_locks.get(path, None)
    if lock is None:
        lock = RLock()
        time_execution_path_locks[path] = lock
    return lock


@contextmanager
def time_execution(identifier: str = None,
                   time_ns_func: TIME_NS_FUNC_TYPE = perf_counter_ns,
                   output: Union[Callable, Path] = print,
                   condition: bool = None,
                   as_seconds: bool = True,
                   decimal_places: Union[int, None] = None,
                   also_pretty: bool = False) -> None:
    if condition is None:
        condition = _check_env_var_condition("TIME_EXECUTION_ENABLED")
    if condition:
        start_time = time_ns_func()
        yield
        full_time = time_ns_func() - start_time
        if as_seconds is True:
            from gidapptools.general_helper.conversion import ns_to_s
            full_time = ns_to_s(full_time, decimal_places=decimal_places)
            unit = 's'
            pretty = "" if also_pretty is False else f" ({seconds2human(full_time)})"
        else:
            unit = 'ns'
            pretty = "" if also_pretty is False else f" ({seconds2human(ns_to_s(full_time))})"
        identifier = '' if identifier is None else identifier

        if isinstance(output, Path):
            with get_time_execution_path_lock(output):
                with output.open('a', encoding='utf-8', errors='ignore') as f:
                    f.write(f"{identifier} took {full_time:f} {unit}{pretty}" + '\n')
        else:
            output(f"{identifier} took {full_time:f} {unit}{pretty}")
    else:
        yield


def time_func(time_ns_func: TIME_NS_FUNC_TYPE = perf_counter_ns,
              output: Callable = print,
              use_qualname: bool = True,
              condition: bool = None,
              as_seconds: bool = True,
              decimal_places: int = None,
              also_pretty: bool = False):

    def _decorator(func):
        func_name = func.__name__ if use_qualname is False else func.__qualname__
        _actual_condition = _check_env_var_condition("TIME_FUNC_ENABLED") if condition is None else condition

        @wraps(func)
        def _wrapped(*args, **kwargs):
            with time_execution(f"executing {func_name!r}", time_ns_func=time_ns_func, output=output, as_seconds=as_seconds, decimal_places=decimal_places, condition=True, also_pretty=also_pretty):
                return func(*args, **kwargs)

        if _actual_condition:
            return _wrapped
        return func

    return _decorator


# def profile(func):
#     """
#     from `https://gist.github.com/pavelpatrin/5a28311061bf7ac55cdd`
#     """

#     @wraps(func)
#     def wrapper(*args, **kwargs):
#         from line_profiler import LineProfiler
#         prof = LineProfiler()
#         try:
#             return prof(func)(*args, **kwargs)
#         finally:
#             prof.print_stats()

#     return wrapper


def profile(func):
    """
    Dummy decorator to be able to leave the `line_profiler`-decorator `@profile` in place,
    even when not line-profiling.
    """
    return func


def get_dummy_profile_decorator_in_globals():

    if os.getenv("LINE_PROFILE_RUNNING", "0") == "1":
        return
    stk = inspect.stack()[1]
    mod = inspect.getmodule(stk[0])
    if mod is not None:
        setattr(mod, "profile", profile)


if __name__ == '__main__':
    pass



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import re
import json
from typing import Any, Union, Literal, Callable, Hashable
from pathlib import Path
from datetime import datetime, timedelta
from threading import RLock

# * Third Party Imports --------------------------------------------------------------------------------->
from yarl import URL

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.types import PATH_TYPE
from gidapptools.gid_config.enums import SpecAttribute
from gidapptools.general_helper.enums import MiscEnum
from gidapptools.general_helper.dict_helper import BaseVisitor, AdvancedDict, KeyPathError, set_by_key_path
from gidapptools.general_helper.string_helper import split_quotes_aware
from gidapptools.general_helper.mixins.file_mixin import FileMixin
from gidapptools.gid_config.conversion.conversion_table import EntryTypus
from gidapptools.gid_config.conversion.extra_base_typus import NonTypeBaseTypus

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


class SpecVisitor(BaseVisitor):
    argument_regex = re.compile(r"(?P<base_type>\w+)\(((?P<sub_arguments>.*)\))?")
    sub_arguments_regex = re.compile(r"(?P<name>\w+)\s*\=\s*(?P<value>.*)")

    def __init__(self, extra_handlers: dict[Hashable, Callable] = None, default_handler: Callable = None, sub_argument_separator: str = ',') -> None:
        super().__init__(extra_handlers=extra_handlers, default_handler=default_handler)
        self.sub_argument_separator = sub_argument_separator

    def visit(self, in_dict: Union["AdvancedDict", dict], key_path: tuple[str], value: Any) -> None:

        key_path = tuple(key_path)

        if key_path[-1] != 'converter':
            return
        value_key = self._modify_value(value)

        handler = self.handlers.get(key_path, self.handlers.get(value_key, self.default_handler))
        if handler is None:
            return
        if isinstance(in_dict, AdvancedDict):
            in_dict.set(key_path, handler(value, self._get_sub_arguments(value)))
        else:
            set_by_key_path(in_dict, key_path, handler(value, self._get_sub_arguments(value)))

    def _modify_value(self, value: Any) -> Any:
        value = super()._modify_value(value)
        try:
            value = self.argument_regex.sub(r"\g<base_type>", value)
        except (AttributeError, TypeError):
            pass

        return value

    def _get_handler_direct(self, value: str) -> Callable:
        return self.handlers.get(value, self._handle_string)

    def _get_sub_arguments(self, value: str, default: dict[str:str] = None) -> dict[str:str]:

        def _get_key_value_from_part(in_part: str) -> dict[str, str]:
            sub_match = self.sub_arguments_regex.match(in_part)
            name = sub_match.group('name')
            value = sub_match.group('value')
            return {name.strip(): value.strip().strip('"' + "'").strip()}

        default = {} if default is None else default
        try:
            match = self.argument_regex.match(value)
            sub_arguments_string = match.groupdict().get("sub_arguments", default)
            sub_arguments = {}
            for part in split_quotes_aware(sub_arguments_string, quote_chars="'", strip_parts=True):
                try:
                    sub_arguments |= _get_key_value_from_part(part)
                except AttributeError:
                    continue
            if not sub_arguments:
                return default
            return sub_arguments
        except AttributeError:
            return default

    def _handle_default(self, value: Any, sub_arguments: dict[str, str]) -> EntryTypus:
        """
        handles all values that other handlers, can't or that raised an error while dispatching to handlers.

        NAMED_VALUE_ARGUMENTS:
            None
        Args:
            value (Any): The nodes value.

        Returns:
            EntryTypus: An `EntryTypus` with only an `original_value` and the base_typus set to `SpecialTypus.DELAYED)
        """
        return EntryTypus(original_value=value)

    def _handle_boolean(self, value: Any, sub_arguments: dict[str, str]) -> EntryTypus:
        """
        NAMED_VALUE_ARGUMENTS:
            None
        Args:
            value (Any): [description]

        Returns:
            EntryTypus: [description]
        """

        return EntryTypus(original_value=value, base_typus=bool)

    def _handle_string(self, value: Any, sub_arguments: dict[str, str] = None) -> EntryTypus:
        """
        NAMED_VALUE_ARGUMENTS:
            None
        Args:
            value (Any): [description]

        Returns:
            EntryTypus: [description]
        """
        sub_arguments = sub_arguments or {}
        if "choices" in sub_arguments:
            sub_arguments["choices"] = [i.strip() for i in sub_arguments["choices"].split('|')]
        return EntryTypus(original_value=value, base_typus=str, named_arguments=sub_arguments)

    def _handle_integer(self, value: Any, sub_arguments: dict[str, str]) -> EntryTypus:
        """
        NAMED_VALUE_ARGUMENTS:
            None
        Args:
            value (Any): [description]

        Returns:
            EntryTypus: [description]
        """
        return EntryTypus(original_value=value, base_typus=int)

    def _handle_float(self, value: Any, sub_arguments: dict[str, str]) -> EntryTypus:
        """
        NAMED_VALUE_ARGUMENTS:
            None
        Args:
            value (Any): [description]

        Returns:
            EntryTypus: [description]
        """
        return EntryTypus(original_value=value, base_typus=float)

    def _handle_bytes(self, value: Any, sub_arguments: dict[str, str]) -> EntryTypus:
        """
        NAMED_VALUE_ARGUMENTS:
            None
        Args:
            value (Any): [description]

        Returns:
            EntryTypus: [description]
        """
        return EntryTypus(original_value=value, base_typus=bytes)

    def _handle_list(self, value: Any, sub_arguments: dict[str, str]) -> EntryTypus:
        """
        Converts the value to `list` with optional sub_type (eg: `list[int]`).

        NAMED_VALUE_ARGUMENTS:
            subtype: The subtype of the list, defaults to `string`, can be any other handled type.
        Args:
            value (Any): [description]

        Returns:
            EntryTypus: [description]
        """
        def _process_subtypus(in_sub_arguments: dict[str, str]) -> None:
            subtypus_string = sub_arguments.pop('sub_typus', 'string')
            subtypus = self._get_handler_direct(subtypus_string)(subtypus_string, {})
            in_sub_arguments['sub_typus'] = subtypus

        def _process_split_char(in_sub_arguments: dict[str, str]) -> None:
            if "split_char" not in in_sub_arguments:
                in_sub_arguments["split_char"] = ','
        _process_subtypus(sub_arguments)
        _process_split_char(sub_arguments)

        return EntryTypus(original_value=value, base_typus=list, named_arguments=sub_arguments)

    def _handle_datetime(self, value: Any, sub_arguments: dict[str, str]) -> EntryTypus:
        """
        [summary]

        NAMED_VALUE_ARGUMENTS:
            None
        Args:
            value (Any): [description]

        Returns:
            EntryTypus: [description]
        """

        return EntryTypus(original_value=value, base_typus=datetime)

    def _handle_path(self, value: Any, sub_arguments: dict[str, str]) -> EntryTypus:
        """
        NAMED_VALUE_ARGUMENTS:
            None
        Args:
            value (Any): [description]

        Returns:
            EntryTypus: [description]
        """

        return EntryTypus(original_value=value, base_typus=Path)

    def _handle_url(self, value: Any, sub_arguments: dict[str, str]) -> EntryTypus:
        """
        NAMED_VALUE_ARGUMENTS:
            None
        Args:
            value (Any): [description]

        Returns:
            EntryTypus: [description]
        """

        return EntryTypus(original_value=value, base_typus=URL)

    def _handle_file_size(self, value: Any, sub_arguments: dict[str, str]) -> EntryTypus:
        """
        NAMED_VALUE_ARGUMENTS:
            None
        Args:
            value (Any): [description]

        Returns:
            EntryTypus: [description]
        """
        return EntryTypus(original_value=value, base_typus=NonTypeBaseTypus.FILE_SIZE)

    def _handle_timedelta(self, value: Any, sub_arguments: dict[str, str]) -> EntryTypus:
        """
        NAMED_VALUE_ARGUMENTS:
            None
        Args:
            value (Any): [description]

        Returns:
            EntryTypus: [description]
        """
        return EntryTypus(original_value=value, base_typus=timedelta)


class SpecData(AdvancedDict):
    default_visitor_class = SpecVisitor
    visit_lock = RLock()

    def __init__(self, visitor: SpecVisitor, **kwargs) -> None:
        self.visitor = visitor
        super().__init__(data=None, **kwargs)

    @property
    def data(self) -> dict:
        with self.visit_lock:
            return super().data

    def _get_section_default(self, section_name: str) -> EntryTypus:

        return self.get([section_name, '__default__', "converter"], str)

    def _get_entry_default(self, section_name: str, entry_key: str) -> Union[str, MiscEnum]:
        return self.get([section_name, entry_key, 'default'], MiscEnum.NOTHING)

    def get_entry_typus(self, section_name: str, entry_key: str) -> EntryTypus:
        try:
            return self[[section_name, entry_key, "converter"]]
        except KeyPathError as error:
            try:
                return self._get_section_default(section_name=section_name)
            except KeyPathError:
                raise error

    def get_description(self, section_name: str, entry_key: str) -> str:
        return self.get(key_path=[section_name, entry_key, SpecAttribute.DESCRIPTION.value], default="")

    def get_spec_attribute(self, section_name: str, entry_key: str, attribute: Union[SpecAttribute, str], default=None) -> Any:
        attribute = SpecAttribute(attribute) if isinstance(attribute, str) else attribute
        return self.get([section_name, entry_key, attribute.value], default=default)

    def set_typus_value(self, section_name: str, entry_key: str, typus_value: str) -> None:
        if section_name not in self:
            self[section_name] = {}
        self[section_name][entry_key] = typus_value

    def _resolve_values(self) -> None:
        self.modify_with_visitor(self.visitor)

    def reload(self) -> None:
        with self.visit_lock:
            self.visitor.reload()
            self._resolve_values()


class SpecFile(FileMixin, SpecData):
    def __init__(self, file_path: PATH_TYPE, visitor: SpecVisitor, changed_parameter: Union[Literal['size'], Literal['file_hash']] = 'size', **kwargs) -> None:
        super().__init__(visitor=visitor, file_path=file_path, changed_parameter=changed_parameter, ** kwargs)
        self._data = None
        self.spec_name = self.file_path.stem.casefold()

    @property
    def data(self) -> dict:
        if self._data is None or self.has_changed is True:
            self.load()
        return self._data

    def reload(self) -> None:
        self.load()

    def load(self) -> None:
        with self.lock:
            self._data = json.loads(self.read())
            super().reload()

    def save(self) -> None:
        with self.lock:
            json_data = json.dumps(self.data, indent=4, sort_keys=False, default=lambda x: x.convert_for_json())
            self.write(json_data)

    def set_typus_value(self, section_name: str, entry_key: str, typus_value: str) -> None:
        super().set_typus_value(section_name=section_name, entry_key=entry_key, typus_value=typus_value)
        self.save()

# class SpecDataFile(SpecData):

#     def __init__(self, in_file: Path, changed_parameter: str = 'size', ensure: bool = True, visitor_class: SpecVisitor = None, **kwargs) -> None:
#         super().__init__(visitor_class=visitor_class, **kwargs)

#         self.file_path = Path(in_file).resolve()
#         self.changed_parameter = changed_parameter
#         self.ensure = ensure
#         self.last_size: int = None
#         self.last_file_hash: str = None
#         self.data = None
#         self.lock = Lock()

#     @property
#     def has_changed(self) -> bool:
#         if self.changed_parameter == 'always':
#             return True
#         if self.changed_parameter == 'both':
#             if any([param is None for param in [self.last_size, self.last_file_hash]] + [self.last_size != self.size, self.last_file_hash != self.last_file_hash]):
#                 return True
#         if self.changed_parameter == 'size':
#             if self.last_size is None or self.size != self.last_size:
#                 return True
#         elif self.changed_parameter == 'file_hash':
#             if self.last_file_hash is None or self.file_hash != self.last_file_hash:
#                 return True
#         return False

#     def get_converter(self, key_path: Union[list[str], str]) -> EntryTypus:
#         with self.lock:
#             if self.data is None or self.has_changed is True:
#                 self.reload(locked=True)
#             return super().get_converter(key_path)

#     @property
#     def size(self) -> int:
#         return self.file_path.stat().st_size

#     @property
#     def file_hash(self) -> str:
#         _file_hash = blake2b()
#         with self.file_path.open('rb') as f:
#             for chunk in f:
#                 _file_hash.update(chunk)
#         return _file_hash.hexdigest()

#     def reload(self, locked: bool = False) -> None:
#         self.load(locked)
#         super().reload()

#     def _json_converter(self, item: Union[EntryTypus, type]) -> str:
#         try:
#             return item.convert_for_json()
#         except AttributeError:
#             return EntryTypus.special_name_conversion_table(type.__name__, type.__name__)

#     def load(self, locked: bool = False):
#         def _load():
#             with self.file_path.open('r', encoding='utf-8', errors='ignore') as f:
#                 return json.load(f)

#         if self.file_path.exists() is False and self.ensure is True:
#             self.write(locked=locked)
#         if locked is False:
#             with self.lock:
#                 self.data = _load()
#         else:
#             self.data = _load()

#     def write(self, locked: bool = False) -> None:
#         def _write():
#             with self.file_path.open('w', encoding='utf-8', errors='ignore') as f:
#                 data = {} if self.data is None else self.data
#                 json.dump(data, f, default=self._json_converter, indent=4, sort_keys=True)

#         if locked is False:
#             with self.lock:
#                 _write()
#         else:
#             _write()


# region[Main_Exec]

if __name__ == '__main__':
    pass
# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from enum import Enum, auto
from pathlib import Path

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


class NonTypeBaseTypus(Enum):
    FILE_SIZE = auto()
    STRING_CHOICE = auto()


# region[Main_Exec]
if __name__ == '__main__':
    pass

# endregion[Main_Exec]



NANOSECONDS_IN_SECOND: int = 1_000_000_000


RAW_STRING_TRUE_VALUES = {'yes',
                          'y',
                          '1',
                          'true',
                          '+'}

RAW_STRING_FALSE_VALUES = {'no',
                           'n',
                           '0',
                           'false',
                           '-'}


STRING_TRUE_VALUES = {str(value).casefold() for value in RAW_STRING_TRUE_VALUES}

STRING_FALSE_VALUES = {str(value).casefold() for value in RAW_STRING_FALSE_VALUES}



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import json
from typing import TYPE_CHECKING, Any, Union, Literal, Callable, Optional
from pathlib import Path

# * Third Party Imports --------------------------------------------------------------------------------->
import attr

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.utility.helper import make_pretty
from gidapptools.gid_config.interface import GidIniConfig
from gidapptools.abstract_classes.abstract_meta_factory import AbstractMetaItem, AbstractMetaFactory

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.meta_data.config_kwargs import ConfigKwargs

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]

CONFIG_SPEC_PAIRS_TYPE = dict[
    str, dict[
        Union[Literal["config_file"], Literal["spec_file"]], Path
    ]
]


@attr.s(auto_attribs=True, auto_detect=True, kw_only=True)
class MetaConfig(AbstractMetaItem):
    config_paths: list[Path] = attr.ib()
    spec_paths: list[Path] = attr.ib()
    config_name_suffix: str = attr.ib(default='_config', converter=lambda x: x.casefold())
    spec_name_suffix: str = attr.ib(default="_configspec", converter=lambda x: x.casefold())
    config_class: type = attr.ib(default=GidIniConfig)
    file_changed_parameter: str = attr.ib(default='size')
    instances: dict[tuple[Path, Path], config_class] = {}

    @property
    def config_spec_pairs(self) -> CONFIG_SPEC_PAIRS_TYPE:
        pairs = {}
        configs = {item.stem.casefold().removesuffix(self.config_name_suffix): item for item in self.config_paths}
        specs = {item.stem.casefold().removesuffix(self.spec_name_suffix): item for item in self.spec_paths}
        for name, path in configs.items():
            pairs[name] = {"config_file": path, "spec_file": specs.get(name, None)}

        return pairs

    def get_config(self, name: str) -> Optional[GidIniConfig]:
        config_spec_pair = self.config_spec_pairs[name.casefold()]
        if (config_spec_pair.get("config_file"), config_spec_pair.get("spec_file")) not in self.instances:
            config = self.config_class(**config_spec_pair, file_changed_parameter=self.file_changed_parameter)
            config.reload()
            self.instances[(config_spec_pair.get("config_file"), config_spec_pair.get("spec_file"))] = config

        return self.instances[(config_spec_pair.get("config_file"), config_spec_pair.get("spec_file"))]

    def to_storager(self, storager: Callable = None) -> None:
        if storager is None:
            return
        storager(self)

    def clean_up(self, **kwargs) -> None:
        pass

    def as_dict(self, pretty: bool = False) -> dict[str, Any]:

        if pretty is True:
            return make_pretty(self)
        return attr.asdict(self)


class MetaConfigFactory(AbstractMetaFactory):
    product_class = MetaConfig
    default_configuration = {}

    def __init__(self, config_kwargs: "ConfigKwargs") -> None:
        super().__init__(config_kwargs=config_kwargs)
        self.config_dir: Path = config_kwargs.get("config_dir")
        self.spec_dir: Path = config_kwargs.get("config_spec_dir")
        self.config_paths: list[Path] = []
        self.spec_paths: list[Path] = []

    def _dev_init(self) -> None:
        self.config_dir: Path = self.config_kwargs.get("code_base_dir").joinpath("dev_temp", "config")
        self.spec_dir: Path = self.config_kwargs.get("code_base_dir").joinpath("dev_temp", "config", "spec")

    def _create_spec_file(self, name: str, content: Union[str, dict]) -> Path:
        self.spec_dir.mkdir(exist_ok=True, parents=True)
        full_path = self.spec_dir.joinpath(name).with_suffix('.json')
        if not isinstance(content, str):
            content = json.dumps(content, default=str, sort_keys=False, indent=4)
        if full_path.exists() is False:
            full_path.write_text(content, encoding='utf-8', errors='ignore')
        return full_path

    def _create_config_file(self, name: str, content: str) -> Path:
        self.config_dir.mkdir(exist_ok=True, parents=True)
        full_path = self.config_dir.joinpath(name).with_suffix('.ini')
        if full_path.exists() is False:
            full_path.write_text(content, encoding='utf-8', errors='ignore')
        return full_path

    def create_files(self, file_to_create: Union[Path, tuple[str, Union[str, dict]]], typus) -> Path:
        if isinstance(file_to_create, Path):
            file_to_create = (file_to_create.name, file_to_create.read_text(encoding='utf-8', errors='ignore'))
        func = self._create_config_file if typus == "config" else self._create_spec_file
        return func(*file_to_create)

    def setup(self) -> None:
        if self.config_kwargs.get('is_dev', False) is True:
            self._dev_init()

        for config_to_create in self.config_kwargs.get("configs_to_create", []):
            self.config_paths.append(self.create_files(config_to_create, "config"))
        for spec_to_create in self.config_kwargs.get("spec_to_create", []):
            self.spec_paths.append(self.create_files(spec_to_create, "spec"))

    def _build(self) -> AbstractMetaItem:
        if self.is_setup is False:
            self.setup()
        kwargs = self.config_kwargs.get_kwargs_for(self.product_class, defaults={"config_paths": self.config_paths, "spec_paths": self.spec_paths})
        return self.product_class(**kwargs)

# region[Main_Exec]


if __name__ == '__main__':
    pass

# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import sys
from typing import TYPE_CHECKING, Union, Optional
from pathlib import Path
from weakref import WeakSet

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import QIcon, QImage, QPixmap, QGuiApplication
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QMainWindow, QMessageBox, QApplication, QSystemTrayIcon

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gidapptools_qt.resources.placeholder import QT_PLACEHOLDER_IMAGE

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gidapptools_qt.resources.resources_helper import PixmapResourceItem

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


class GidQtApplication(QApplication):

    def __init__(self,
                 argvs: list[str] = sys.argv,
                 icon: Union["PixmapResourceItem", QPixmap, QImage, str, QIcon] = None):
        super().__init__(self.argv_hook(argvs))
        self.main_window: QMainWindow = None
        self.sys_tray: QSystemTrayIcon = None
        self.icon = self._icon_conversion(icon)
        self.extra_windows = WeakSet()

    @classmethod
    def with_pre_flags(cls,
                       argvs: list[str] = sys.argv,
                       icon: Union["PixmapResourceItem", QPixmap, QImage, str, QIcon] = None,
                       pre_flags: dict[Qt.ApplicationAttribute:bool] = None,
                       desktop_settings_aware: bool = True):
        QGuiApplication.setDesktopSettingsAware(desktop_settings_aware)
        for flag, value in pre_flags.items():
            cls.setAttribute(flag, value)
        return cls(argvs=argvs, icon=icon)

    def setup(self) -> "GidQtApplication":
        self.setWindowIcon(self.icon)
        return self

    def argv_hook(self, argvs: list[str]) -> list[str]:
        return argvs

    @staticmethod
    def _icon_conversion(icon: Union["PixmapResourceItem", QPixmap, QImage, str, QIcon] = None) -> Optional[QIcon]:
        if icon is None:
            return QT_PLACEHOLDER_IMAGE.icon

        if isinstance(icon, QIcon):
            return icon

        if isinstance(icon, (QPixmap, QImage, str)):
            return QIcon(icon)

        return icon.get_as_icon()

    def show_about_qt(self) -> None:
        self.aboutQt()

    def _get_about_text(self) -> str:
        text_parts = {"Name": self.applicationDisplayName(),
                      "Author": self.organizationName(),
                      "Link": f'<a href="{self.organizationDomain()}">{self.organizationDomain()}</a>',
                      "Version": self.applicationVersion()}

        return '<br>'.join(f"<b>{k:<20}:</b>{v:>50}" for k, v in text_parts.items())

    def show_about(self) -> None:
        title = f"About {self.applicationDisplayName()}"
        text = self._get_about_text()
        QMessageBox.about(self.main_window, title, text)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.applicationDisplayName()!r})"


# region[Main_Exec]
if __name__ == '__main__':
    app = GidQtApplication(sys.argv)
    m = QMainWindow()
    m.show()
    app.exec_()

# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from abc import ABC, abstractmethod
from typing import Any, Union, Iterable, Optional
from pathlib import Path
from colorsys import hls_to_rgb, hsv_to_rgb, rgb_to_hls, rgb_to_hsv
from functools import cached_property

# * Third Party Imports --------------------------------------------------------------------------------->
import attr

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.color.preset_colors.web_colors import get_webcolors_data

try:
    from PySide6.QtGui import QColor
    PYSIDE6_AVAILABLE = True
except ImportError:
    PYSIDE6_AVAILABLE = False

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]

INT_OR_FLOAT = Union[int, float]

COLOR_FLOAT_TYPE = Union[tuple[float, float, float], tuple[float, float, float, float]]

COLOR_INT_TYPE = Union[tuple[int, int, int], tuple[int, int, int, float]]


def _validate_color_float(instance: object, attribute: attr.Attribute, value: Optional[float]) -> None:

    if 0.0 > value > 1.0:

        raise ValueError(f"{attribute.name!r}-value can only be between 0 and 1, not {value!r}.")


def _alpha_converter(alpha: Union[int, float] = None) -> float:
    if alpha is None:
        return 1.0

    return float(alpha)


class BaseColor(ABC):
    float_round_n: int = 4

    def __iter__(self):
        return (getattr(self, field.name) for field in attr.fields(self.__class__) if field.metadata.get("is_core_color", False) is True)

    @abstractmethod
    def as_rgb_float(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        ...

    @abstractmethod
    def as_rgb_int(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_INT_TYPE:
        ...

    @abstractmethod
    def as_hls(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        ...

    @abstractmethod
    def as_hsv(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        ...

    @cached_property
    def qcolor(self) -> Optional["QColor"]:
        if PYSIDE6_AVAILABLE:
            return QColor(*self.as_rgb_int(True, True))

    def to_q_style_rgba_string(self,) -> str:
        values = (str(i) for i in self.as_rgb_int(alpha_as_int=True))
        return "rgba(" + ', '.join(values) + ')'

    def __repr__(self) -> str:
        class_name = self.__class__.__name__.removesuffix("Color")
        _out = f"{class_name}(" + ', '.join(repr(x) for x in self)
        if self.name is not None:
            _out += f', name={self.name!r}'

        if self.aliases:
            _out += f", aliases={tuple(self.aliases)!r}"
        return _out + ')'

    def with_alpha(self, alpha: Union[int, float]) -> "BaseColor":
        if isinstance(alpha, int) or alpha > 1:
            alpha = round(alpha / 255, self.float_round_n)
        return self.__class__(*list(self)[:-1], alpha)


@attr.s(auto_attribs=True, auto_detect=True, slots=True, weakref_slot=True, frozen=True, order=True, repr=False)
class RGBColor(BaseColor):
    red: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})
    green: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})
    blue: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})

    alpha: float = attr.ib(default=1.0, converter=_alpha_converter, validator=_validate_color_float, metadata={'is_core_color': True})

    name: str = attr.ib(default=None)
    aliases: Iterable[str] = attr.ib(factory=set, converter=attr.converters.default_if_none(factory=set))

    def as_rgb_float(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = list(self)[:-1]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_rgb_int(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_INT_TYPE:
        _out = [int(sub_col * 255) for sub_col in list(self)[:-1]]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_hls(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = [round(col, self.float_round_n) for col in rgb_to_hls(*list(self)[:-1])]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return _out

    def as_hsv(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = [round(col, self.float_round_n) for col in rgb_to_hsv(*list(self)[:-1])]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return _out


@attr.s(auto_attribs=True, auto_detect=True, slots=True, weakref_slot=True, frozen=True, order=True, repr=False)
class HLSColor(BaseColor):
    hue: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})
    saturation: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})
    lightness: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})

    alpha: float = attr.ib(default=1.0, converter=_alpha_converter, validator=_validate_color_float, metadata={'is_core_color': True})

    name: str = attr.ib(default=None)
    aliases: Iterable[str] = attr.ib(factory=set, converter=attr.converters.default_if_none(factory=set))

    def as_hls(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = list(self)[:-1]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_rgb_float(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = [round(col, self.float_round_n) for col in hls_to_rgb(*list(self)[:-1])]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_rgb_int(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_INT_TYPE:
        _out = [int(sub_col * 255) for sub_col in hls_to_rgb(*list(self)[:-1])]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_hsv(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = [round(col, self.float_round_n) for col in rgb_to_hsv(*hls_to_rgb(*list(self)[:-1]))]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return _out


@attr.s(auto_attribs=True, auto_detect=True, slots=True, weakref_slot=True, frozen=True, order=True, repr=False)
class HSVColor(BaseColor):
    hue: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})
    saturation: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})
    value: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})

    alpha: float = attr.ib(default=1.0, converter=_alpha_converter, validator=attr, metadata={'is_core_color': True})

    name: str = attr.ib(default=None)
    aliases: Iterable[str] = attr.ib(factory=set, converter=attr.converters.default_if_none(factory=set))

    def as_hsv(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = list(self)[:-1]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_hls(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = [round(col, self.float_round_n) for col in rgb_to_hls(*hsv_to_rgb(*list(self)[:-1]))]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return _out

    def as_rgb_float(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = [round(col, self.float_round_n) for col in hsv_to_rgb(*list(self)[:-1])]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_rgb_int(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_INT_TYPE:
        _out = [int(sub_col * 255) for sub_col in hsv_to_rgb(*list(self)[:-1])]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)


color_class_map: dict[str, type[BaseColor]] = {"rgb": RGBColor,
                                               "hls": HLSColor,
                                               "hsv": HSVColor}


class ColorRegistry:

    def __init__(self) -> None:
        self.colors_by_name: dict[str:RGBColor] = {}
        self.colors_by_rgba_value: dict[tuple[float, float, float, float]:RGBColor] = {}

    def _add_color(self, color: RGBColor) -> None:
        if color.name:
            self.colors_by_name[color.name.casefold()] = color
        for alias in color.aliases:
            self.colors_by_name[alias.casefold()] = color

        self.colors_by_rgba_value[color.as_rgb_float(True, False)] = color

    def get_color(self, r: float, g: float, b: float, alpha: float = 1.0, name: str = None, aliases: Iterable[str] = None) -> RGBColor:
        if name is not None:
            if color := self.colors_by_name.get(name.casefold()):
                return color
        if color := self.colors_by_rgba_value.get((r, g, b, alpha)):
            return color

        color = RGBColor(r, g, b, alpha, name, aliases)
        self._add_color(color)
        return color

    def color_factory(self, r: Union[int, float], g: Union[int, float], b: Union[int, float], alpha: float = 1.0, name: str = None, aliases: Iterable[str] = None):

        if any(sub_v > 1 for sub_v in (r, g, b)) or all(isinstance(sub_v, int) for sub_v in (r, g, b)):
            values = tuple(round(sub_v / 255, BaseColor.float_round_n) for sub_v in (r, g, b))

        if alpha is not None and alpha > 1:
            alpha = round(alpha / 255, BaseColor.float_round_n)
        return self.get_color(*values, alpha=alpha, name=name, aliases=aliases)

    def __call__(self, r: Union[int, float], g: Union[int, float], b: Union[int, float], alpha: float = 1.0, name: str = None, aliases: Iterable[str] = None) -> Any:
        return self.color_factory(r, g, b, alpha, name, aliases)

    def get_color_by_name(self, name: str) -> "RGBColor":
        return self.colors_by_name[name.casefold()]


Color = ColorRegistry()
for p_color in get_webcolors_data():
    Color(**p_color)

# region[Main_Exec]
if __name__ == '__main__':
    pass
# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import sys
import platform
from typing import Any, Callable, Optional
from pathlib import Path
from datetime import datetime, timezone
from functools import cached_property

# * Third Party Imports --------------------------------------------------------------------------------->
import attr
from yarl import URL
from tzlocal import get_localzone, reload_localzone

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.utility.enums import OperatingSystem
from gidapptools.utility.helper import utc_now, make_pretty, memory_in_use
from gidapptools.utility._debug_tools import dprint
from gidapptools.general_helper.general import is_frozen
from gidapptools.general_helper.date_time import DatetimeFmt
from gidapptools.general_helper.conversion import bytes2human
from gidapptools.general_helper.string_helper import StringCase, StringCaseConverter
from gidapptools.abstract_classes.abstract_meta_item import AbstractMetaItem

reload_localzone()
print = dprint

# endregion[Imports]

# region [TODO]

# - Make into a class

# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()


# endregion[Constants]

def url_converter(in_url: str) -> Optional[URL]:
    if in_url is None:
        return in_url
    return URL(in_url)


@attr.s(auto_attribs=True, auto_detect=True, kw_only=True, frozen=True)
class MetaInfo(AbstractMetaItem):
    app_name: str = attr.ib(default=None)
    app_author: str = attr.ib(default=None)
    version: str = attr.ib(default=None)
    url: URL = attr.ib(converter=url_converter, default=None)
    pid: int = attr.ib(factory=os.getpid)
    os: OperatingSystem = attr.ib(factory=OperatingSystem.determine_operating_system)
    os_release: str = attr.ib(factory=platform.release)
    python_version: str = attr.ib(factory=platform.python_version)
    started_at: datetime = attr.ib(factory=utc_now)
    base_mem_use: int = attr.ib(default=memory_in_use())
    is_dev: bool = attr.ib(default=None, converter=attr.converters.default_if_none(False))
    is_gui: bool = attr.ib(default=None, converter=attr.converters.default_if_none(False))
    local_tz: timezone = attr.ib(default=get_localzone())

    @cached_property
    def is_frozen(self) -> bool:
        return is_frozen()

    @cached_property
    def frozen_folder_path(self) -> Optional[Path]:
        if self.is_frozen is True:
            return Path(sys._MEIPASS)

    @classmethod
    @property
    def __default_configuration__(cls) -> dict[str, Any]:
        default_configuration = {}
        return default_configuration

    @property
    def pretty_is_dev(self) -> str:
        return "Yes" if self.is_dev else "No"

    @property
    def pretty_app_name(self) -> str:
        return StringCaseConverter.convert_to(self.app_name, StringCase.TITLE)

    @property
    def pretty_app_author(self) -> str:
        return StringCaseConverter.convert_to(self.app_author, StringCase.TITLE)

    @property
    def pretty_base_mem_use(self) -> str:
        return bytes2human(self.base_mem_use)

    @property
    def pretty_started_at(self) -> str:
        return DatetimeFmt.STANDARD.strf(self.started_at)

    def as_dict(self, pretty: bool = False) -> dict[str, Any]:

        if pretty is True:
            return make_pretty(self)
        return attr.asdict(self)

    def to_storager(self, storager: Callable = None) -> None:
        if storager is None:
            return
        storager(self)

    def clean_up(self, **kwargs) -> None:
        pass


    # region[Main_Exec]
if __name__ == '__main__':
    pass

# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
from glob import iglob
from typing import TYPE_CHECKING, Union, Optional, Generator
from pathlib import Path
from contextlib import contextmanager
from concurrent.futures import ThreadPoolExecutor

# * Third Party Imports --------------------------------------------------------------------------------->
from psutil import disk_partitions

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.types import PATH_TYPE

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


def get_all_drives(also_non_physical: bool = False) -> tuple[Path]:
    return tuple(Path(drive.mountpoint) for drive in disk_partitions(all=also_non_physical))


def find_file(file_name: str, return_first: bool = True, case_sensitive: bool = False) -> Optional[Union[Path, tuple[Path]]]:
    if case_sensitive is False:
        file_name = file_name.casefold()
    _out = []
    for drive in get_all_drives():
        for dirname, folderlist, filelist in os.walk(drive):
            for file in filelist:
                if file.casefold() == file_name:
                    file_path = Path(dirname, file)
                    if return_first is True:
                        return file_path
                    else:
                        _out.append(file_path)
    if _out != []:
        return tuple(_out)
    return None


def find_file_alternative(file_name: str, return_first: bool = True) -> Optional[Union[Path, tuple[Path]]]:
    def _helper(args) -> Generator[Path, None, None]:
        _drive = args[0]
        _file_name = args[1]
        for _found_file in iglob(f"{_drive.as_posix().rstrip('/')}/**/{_file_name}", recursive=True):
            yield Path(_found_file)
    _out = []
    with ThreadPoolExecutor(3) as pool:
        for gen in pool.map(_helper, ((drive, file_name) for drive in get_all_drives())):
            for file in gen:
                if return_first is True:
                    return file
                else:
                    _out.append(file)
    if _out != []:
        return tuple(_out)
    return None


@contextmanager
def change_cwd(target_cwd: "PATH_TYPE"):
    old_cwd = Path.cwd()
    new_cwd = Path(target_cwd)
    if new_cwd.is_dir() is False:
        raise FileNotFoundError(f"The target_cwd({new_cwd.as_posix()!r}) either does not exist or is a file and not a directory.")
    os.chdir(new_cwd)
    yield
    os.chdir(old_cwd)


# region[Main_Exec]
if __name__ == '__main__':
    pass
# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from enum import Enum
from pathlib import Path
from datetime import datetime, timezone, timedelta
from functools import total_ordering

# * Third Party Imports --------------------------------------------------------------------------------->
import attr

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import DateTimeFrameTimezoneError

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


# def seconds2human()


class DatetimeFmt(str, Enum):
    STANDARD = "%Y-%m-%d %H:%M:%S"
    FILE = "%Y-%m-%d_%H-%M-%S"
    LOCAL = "%x %X"

    STANDARD_TZ = "%Y-%m-%d %H:%M:%S %Z"
    FILE_TZ = "%Y-%m-%d_%H-%M-%S_%Z"
    LOCAL_TZ = "%x %X %Z"

    def strf(self, date_time: datetime) -> str:
        non_tz_fmt = self if not self.name.endswith('_TZ') else getattr(self, self.name.removesuffix('_TZ'))
        tz_fmt = self if self.name.endswith('_TZ') else getattr(self, self.name + '_TZ')
        if date_time.tzinfo is None:
            return date_time.strftime(non_tz_fmt)
        return date_time.strftime(tz_fmt)


# class NamedTimezone:
#     LOCAL = get_localzone()
#     UTC = timezone.utc
#     GMT = timezone(offset=timedelta(hours=0), name='GMT')
#     CET = pytz.timezone('CET')
#     CST = pytz.timezone('America/Winnipeg')

#     @classmethod
#     def get(cls, name: str, default: Any = None) -> Optional[tzinfo]:
#         mod_name = name.upper().strip()
#         return getattr(cls, mod_name, default)


def get_aware_now(tz: timezone) -> datetime:
    return datetime.now(tz=tz)


def get_utc_now() -> datetime:
    return get_aware_now(tz=timezone.utc)


def _validate_date_time_frame_tzinfo(instance: "DateTimeFrame", attribute: attr.Attribute, value: datetime):
    if instance.start.tzinfo is None or instance.end.tzinfo is None:
        raise DateTimeFrameTimezoneError(instance, instance.start.tzinfo, instance.end.tzinfo, 'start time and end time need to be timezone aware')
    if instance.start.tzinfo != instance.end.tzinfo:
        raise DateTimeFrameTimezoneError(instance, instance.start.tzinfo, instance.end.tzinfo, 'start time and end time do not have the same timezone')


@attr.s(auto_attribs=True, auto_detect=True, frozen=True)
@total_ordering
class DateTimeFrame:
    start: datetime = attr.ib(validator=_validate_date_time_frame_tzinfo)
    end: datetime = attr.ib(validator=_validate_date_time_frame_tzinfo)

    @property
    def delta(self) -> timedelta:
        return self.end - self.start

    @property
    def tzinfo(self) -> timezone:
        return self.start.tzinfo

    def __eq__(self, other: object) -> bool:
        if isinstance(other, datetime):
            return self.start <= other <= self.end
        if isinstance(other, self.__class__):
            return self.start == other.start and self.end == other.end
        if isinstance(other, timedelta):
            return self.delta == other
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, datetime):
            return self.start < other
        if isinstance(other, self.__class__):
            return self.end < other.start
        if isinstance(other, timedelta):
            return self.delta < other
        return NotImplemented

    def __contains__(self, other: object) -> bool:
        if isinstance(other, datetime):
            return self.start <= other <= self.end
        return NotImplemented

    def __str__(self) -> str:
        return f"{self.start.isoformat()} until {self.end.isoformat()}"

    def __hash__(self) -> int:
        return hash(self.start) + hash(self.end) + hash(self.delta)
# region[Main_Exec]


if __name__ == '__main__':
    pass

# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import re
import inspect
from typing import Union, Literal, Mapping, Callable, Iterable
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.enums import StringCase

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]

# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()


# endregion[Constants]

STRING_CASE_FUNC_TYPE = Callable[[Iterable[str]], str]


class StringCaseConverter:
    SNAKE = StringCase.SNAKE
    SCREAMING_SNAKE = StringCase.SCREAMING_SNAKE
    CAMEL = StringCase.CAMEL
    PASCAL = StringCase.PASCAL
    KEBAP = StringCase.KEBAP
    SPLIT = StringCase.SPLIT
    CLASS = StringCase.CLASS
    TITLE = StringCase.TITLE

    split_pascal_case_regex = re.compile(r"(?<!\_)(\B[A-Z])")
    snake_case_to_pascal_case_regex = re.compile(r"(_|^)(\w)")
    _word_list_split_chars = {'-', '_', ' '}
    _word_list_split_regex = re.compile(r'|'.join(list(_word_list_split_chars) + [r"(?=[A-Z])"]))

    _dispatch_table: dict[str, STRING_CASE_FUNC_TYPE] = None

    @classmethod
    @property
    def dispatch_table(cls) -> dict[StringCase, STRING_CASE_FUNC_TYPE]:
        if cls._dispatch_table is None:
            cls._dispatch_table = {}
            for meth_name, meth_obj in inspect.getmembers(cls):
                if meth_name.startswith('_to_') and meth_name.endswith('_case'):
                    key_name = meth_name.removeprefix('_to_').removesuffix('_case')
                    cls._dispatch_table[StringCase(key_name)] = meth_obj
        return cls._dispatch_table

    @staticmethod
    def _snake_to_pascal_replacer(match: re.Match) -> str:
        return match.group(2).upper()

    @classmethod
    def _to_word_list(cls, in_string: str) -> Iterable[str]:
        in_string = in_string.strip()
        if in_string.isupper():
            in_string = in_string.casefold()

        return [word.casefold() for word in cls._word_list_split_regex.split(in_string) if word]

    @staticmethod
    def _to_snake_case(word_list: Iterable[str]) -> str:
        return '_'.join(word_list).casefold()

    @staticmethod
    def _to_camel_case(word_list: Iterable[str]) -> str:
        return word_list[0].casefold() + ''.join(item.title() for item in word_list[1:])

    @staticmethod
    def _to_pascal_case(word_list: Iterable[str]) -> str:
        return ''.join(item.title() for item in word_list)

    @staticmethod
    def _to_kebap_case(word_list: Iterable[str]) -> str:
        return '-'.join(word_list)

    @staticmethod
    def _to_screaming_snake_case(word_list: Iterable[str]) -> str:
        return '_'.join(word_list).upper()

    @staticmethod
    def _to_split_case(word_list: Iterable[str]) -> str:
        return ' '.join(word_list)

    @staticmethod
    def _to_title_case(word_list: Iterable[str]) -> str:
        return ' '.join(word.title() for word in word_list)

    @staticmethod
    def _to_upper_case(word_list: Iterable[str]) -> str:
        return ' '.join(word.upper() for word in word_list)

    @classmethod
    def convert_to(cls, in_string: str, target_case: Union[str, StringCase]) -> str:
        target_case = StringCase(target_case) if isinstance(target_case, str) else target_case
        word_list = cls._to_word_list(in_string)
        return cls.dispatch_table.get(target_case)(word_list)

    @classmethod
    def to_snake_case(cls, in_string: str) -> str:
        return cls.convert_to(in_string=in_string, target_case=StringCase.SNAKE)


def replace_by_dict(in_string: str, in_dict: dict[str, str]) -> str:
    mod_string = in_string
    for key, value in in_dict.items():
        mod_string = mod_string.replace(key, value)
    return mod_string


def extract_by_map(in_string: str, extract_data: Union[Iterable[str], Mapping[str, str]]) -> Iterable[str]:
    parts = []
    re_pattern = re.compile(r'|'.join(extract_data))
    for match in re_pattern.finditer(in_string):
        matched_str = match.group()
        parts.append(matched_str)
    if isinstance(extract_data, Mapping):
        return [extract_data.get(part) for part in parts]
    return parts


SPACE_CLEANING_REGEX = re.compile(r" +")
NEWLINE_CLEANING_REGEX = re.compile(r"\n+")


def clean_whitespace(in_text: str, replace_newline: bool = False) -> str:
    cleaned_text = SPACE_CLEANING_REGEX.sub(' ', in_text)
    if replace_newline is True:
        cleaned_text = NEWLINE_CLEANING_REGEX.sub(' ', cleaned_text)
    return cleaned_text


def shorten_string(in_text: str, max_length: int, shorten_side: Literal["right", "left"] = "right", placeholder: str = '...', clean_before: bool = True, ensure_space_around_placeholder: bool = False, split_on: str = '\s|\n') -> str:
    max_length = int(max_length)
    if shorten_side.casefold() not in {"left", "right"}:
        raise ValueError(shorten_side)

    if clean_before is True:
        in_text = clean_whitespace(in_text, replace_newline=False)

    if len(in_text) <= max_length:
        return in_text

    if ensure_space_around_placeholder is True:
        placeholder = f" {placeholder.strip()}" if shorten_side == "right" else f"{placeholder.strip()} "

    max_length = max_length - len(placeholder)

    new_text = in_text[:max_length] if shorten_side == 'right' else in_text[-max_length:]
    find_regex = re.compile(split_on)
    last_space_position = list(find_regex.finditer(new_text))

    return new_text[:last_space_position[-1].span()[0]].strip() + placeholder if shorten_side == 'right' else placeholder + new_text[last_space_position[0].span()[0]:].strip()


def split_quotes_aware(text: str, split_chars: Iterable[str] = None, quote_chars: Iterable[str] = None, strip_parts: bool = True) -> list[str]:
    """
    Splits a string on but not if the separator char is inside of quotes.



    Args:
        text (str): The string to split.
        split_chars (Iterable[str], optional): The characters to split on. Defaults to `,`.
        quote_chars (Iterable[str], optional): The quote chars that should be considered real quotes. Defaults to `"` and `'`.
        strip_parts (bool, optional): If each found substrin should be striped of preceding and trailing whitespace in the result. Defaults to True.

    Returns:
        list[str]: The found sub-parts.
    """
    split_chars = {','} if split_chars is None else set(split_chars)
    quote_chars = {"'", '"'} if quote_chars is None else set(quote_chars)
    parts = []
    temp_chars = []
    inside_quotes: str = None

    def _add_part():
        nonlocal parts
        nonlocal temp_chars
        part = ''.join(temp_chars)
        if strip_parts is True:
            part = part.strip()
        for quote_char in quote_chars:
            if part.startswith(quote_char) and part.endswith(quote_char):
                part = part.strip(quote_char)
        if part:
            parts.append(part)
        temp_chars.clear()

    for char in text:
        if char in split_chars and inside_quotes is None:
            _add_part()
        else:
            temp_chars.append(char)

            if char in quote_chars and inside_quotes is None:
                inside_quotes = char
            elif char in quote_chars and inside_quotes == char:
                inside_quotes = None

    if temp_chars:
        _add_part()

    return parts


def make_attribute_name(in_string: str) -> str:

    # Remove invalid characters
    in_string = re.sub(r'-', '_', in_string)
    in_string = re.sub('[^0-9a-zA-Z_]', '', in_string)

    # Remove leading characters until we find a letter or underscore
    in_string = re.sub('^[^a-zA-Z_]+', '', in_string)

    return in_string.casefold()

# region[Main_Exec]


if __name__ == '__main__':
    pass

# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Optional
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import QIcon, QAction
from PySide6.QtWidgets import QMenu, QWidget, QMenuBar, QMessageBox, QApplication

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


class BaseMenuBar(QMenuBar):

    def __init__(self, parent: Optional[QWidget] = None, auto_connect_standard_actions: bool = True) -> None:
        super().__init__(parent=parent)
        self.auto_connect_standard_actions = auto_connect_standard_actions
        self.about_dialog: QMessageBox = None
        self.setup_menus()

    def setup_default_menus(self) -> None:
        self.file_menu = self.add_new_menu("File")
        self.exit_action = self.add_new_action(self.file_menu, "Exit")

        self.edit_menu = self.add_new_menu("Edit")

        self.view_menu = self.add_new_menu("View")
        self.windows_menu = self.add_new_menu("Windows", parent_menu=self.view_menu)

        self.settings_menu = self.add_new_menu("Settings")

        self.help_menu = self.add_new_menu("Help")
        self.help_menu.addSeparator()
        self.about_action = self.add_new_action(self.help_menu, "About")
        self.about_qt_action = self.add_new_action(self.help_menu, "About Qt®")

        if self.auto_connect_standard_actions is True:
            self.about_action.triggered.connect(self.app.show_about)
            self.about_qt_action.triggered.connect(self.app.show_about_qt)
            if self.parent() is not None:
                self.exit_action.triggered.connect(self.parent().close)

    @property
    def app(self) -> QApplication:
        return QApplication.instance()

    def setup_menus(self) -> None:
        self.setup_default_menus()

    def add_new_menu(self, menu_title: str, icon: QIcon = None, add_before=None, parent_menu: QMenu = None) -> QMenu:
        menu = QMenu(self)
        menu.setTitle(menu_title)
        if icon is not None:
            menu.setIcon(icon)
        target_menu = parent_menu or self
        if add_before is not None:
            if isinstance(add_before, QMenu):
                add_before = add_before.menuAction()
            target_menu.insertMenu(add_before, menu)
        else:
            target_menu.addMenu(menu)
        return menu

    def add_new_action(self, menu: QMenu, action_name: str, action_title: str = None, add_before=None, action_klass: type[QAction] = QAction):
        action_title = action_title or action_name
        action_name = action_name.casefold().replace(' ', '_')
        action = action_klass(parent=menu)

        action.setText(action_title)
        if add_before is not None:
            if isinstance(add_before, QMenu):
                add_before = add_before.menuAction()
            menu.insertAction(add_before, action)
        else:
            menu.addAction(action)

        if not hasattr(menu, action_name + '_action'):
            setattr(menu, action_name + '_action', action)
        return action

    def add_action(self, menu: QMenu, action: QAction, add_before=None):

        if add_before is not None:
            if isinstance(add_before, QMenu):
                add_before = add_before.menuAction()
            menu.insertAction(add_before, action)
        else:
            menu.addAction(action)

        if not hasattr(menu, action.text() + '_action'):
            setattr(menu, action.text() + '_action', action)
        return action


# region[Main_Exec]
if __name__ == '__main__':
    pass

# endregion[Main_Exec]



"""
WiP
"""

from gidapptools.meta_data.interface import setup_meta_data, get_meta_config, get_meta_info, get_meta_item, get_meta_paths
from gidapptools.gid_logger.logger import get_main_logger, get_logger, get_main_logger_with_file_logging


__version__ = "0.1.14"


from pathlib import Path
THIS_FILE_DIR = Path(__file__).resolve().parent



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Any
from pathlib import Path
from threading import Event

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


class BlockingEvent(Event):

    def __init__(self) -> None:
        super().__init__()
        self.set()

    def __enter__(self) -> None:
        self.clear()

    def __exit__(self, exception_type: type = None, exception_value: BaseException = None, traceback: Any = None) -> None:
        self.set()


# region[Main_Exec]
if __name__ == '__main__':
    pass

# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Any, Union, Mapping, Callable, Hashable
from pathlib import Path
from datetime import datetime, timedelta
from functools import partial

# * Third Party Imports --------------------------------------------------------------------------------->
from yarl import URL

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import ValueValidationError
from gidapptools.general_helper.enums import MiscEnum
from gidapptools.gid_config.parser.tokens import Entry
from gidapptools.general_helper.conversion import bytes2human, human2bytes, str_to_bool, seconds2human, human2timedelta
from gidapptools.general_helper.dispatch_table import BaseDispatchTable
from gidapptools.gid_config.conversion.entry_typus_item import EntryTypus
from gidapptools.gid_config.conversion.extra_base_typus import NonTypeBaseTypus

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]

# Function for conversion
#
# def conversion_function(value: Any, *other_arguments, **named_arguments)->Any:
#     ...


class ValueEncoder:

    def __init__(self, datetime_fmt: str = "isoformat") -> None:
        self.datetime_fmt = datetime_fmt

    def _encode_list(self, value: list) -> str:
        return ', '.join(self.encode(item) for item in value)

    def _encode_datetime(self, value: datetime) -> str:
        if hasattr(value, self.datetime_fmt):
            return self.encode(getattr(value, self.datetime_fmt)())
        else:
            return value.strftime(self.datetime_fmt)

    def _encode_path(self, value: Path) -> str:
        return value.as_posix()

    def _encode_url(self, value: URL) -> str:
        return value.human_repr()

    def encode(self, value: Any) -> str:
        if isinstance(value, str):
            return value
        sub_encoder = {int: str,
                       float: str,
                       bool: str,
                       list: self._encode_list,
                       datetime: self._encode_datetime,
                       Path: self._encode_path,
                       URL: self._encode_url}
        for path_subclass in Path.__subclasses__():
            sub_encoder[path_subclass] = self._encode_path

        return sub_encoder.get(type(value), str)(value)

    def __call__(self, value: Any) -> str:
        return self.encode(value=value)


class ConfigValueConversionTable(BaseDispatchTable):

    def __init__(self, extra_dispatch: Mapping[Hashable, Callable] = None) -> None:
        super().__init__(extra_dispatch=extra_dispatch)

    # pylint: disable=no-self-use
    # pylint: disable=unused-argument
    @BaseDispatchTable.mark(MiscEnum.DEFAULT)
    def _default(self, value: str, mode: str = 'decode', **named_arguments) -> str:
        if mode == "decode":
            return value
        elif mode == "encode":
            return value

    @BaseDispatchTable.mark(str)
    def _string(self, value: str, mode: str = 'decode', **named_arguments) -> str:
        def _validate(_value: str, _named_arguments: dict[str, Any]):
            if "choices" in _named_arguments and value not in named_arguments["choices"]:
                raise ValueValidationError(config_value=value, base_typus=str, validation_description=f"value needs to be one of {_named_arguments['choices']!r}")
        if mode == "decode":
            _validate(value, named_arguments)
            return value
        elif mode == "encode":
            _validate(value, named_arguments)
            return value

    @BaseDispatchTable.mark(int)
    def _integer(self, value: str, mode: str = 'decode', **named_arguments) -> int:

        if mode == "decode":
            return int(value)
        elif mode == "encode":
            return str(value)

    @BaseDispatchTable.mark(float)
    def _float(self, value: str, mode: str = 'decode', **named_arguments) -> float:
        if mode == "decode":
            return float(value)
        elif mode == "encode":
            return str(value)

    @BaseDispatchTable.mark(bool)
    def _boolean(self, value: str, mode: str = 'decode', **named_arguments) -> bool:

        if mode == "decode":
            string_value = str(value)
            new_value = str_to_bool(string_value)
        elif mode == "encode":
            new_value = str(value)

        return new_value

    @BaseDispatchTable.mark(list)
    def _list(self, value: str, mode: str = 'decode', **named_arguments) -> list[Any]:
        if mode == "decode":
            sub_typus = named_arguments.get('sub_typus')
            split_char = named_arguments.get('split_char', ',')
            return [self.convert(entry=item.strip(), typus=sub_typus) for item in value.split(split_char) if item]
        elif mode == "encode":
            return ', '.join(self.encode(item) for item in value)

    @BaseDispatchTable.mark(datetime)
    def _datetime(self, value: str, mode: str = 'decode', **named_arguments) -> datetime:
        if mode == "decode":
            return datetime.fromisoformat(value)
        elif mode == "encode":
            return value.isoformat()

    @BaseDispatchTable.mark(Path, aliases=Path.__subclasses__())
    def _path(self, value: str, mode: str = 'decode', **named_arguments) -> Path:
        if mode == "decode":
            path = Path(value)

            return path
        elif mode == "encode":
            if isinstance(value, str):
                value = Path(value)
            return value.as_posix()

    @BaseDispatchTable.mark(URL)
    def _url(self, value: str, mode: str = 'decode', **named_arguments) -> URL:
        if mode == "decode":
            return URL(value)
        elif mode == "encode":
            if isinstance(value, str):
                return URL(value)
            return value.human_repr()

    @BaseDispatchTable.mark(NonTypeBaseTypus.FILE_SIZE)
    def _file_size(self, value: str, mode: str = 'decode', **named_arguments) -> int:
        if mode == "decode":

            if value.isnumeric() is True:
                return int(value)

            return human2bytes(value)
        elif mode == "encode":

            if isinstance(value, str) and value.isnumeric() is True:
                value = int(value)
            if isinstance(value, int):
                return bytes2human(value)
            return value

    @BaseDispatchTable.mark(timedelta)
    def _timedelta(self, value: str, mode: str = 'decode', **named_arguments) -> timedelta:
        if mode == "decode":
            return human2timedelta(value, default=None)
        elif mode == "encode":
            if isinstance(value, (timedelta, int, float)):
                return seconds2human(value)
            return value

    def _convert_by_type(self, entry: "Entry", typus: type) -> Any:
        converter = self.get(typus)
        value = entry.value if isinstance(entry, Entry) else entry
        return converter(value=value)

    def _convert_by_entry_typus(self, entry: "Entry", typus: EntryTypus) -> Any:
        converter = self.get(typus.base_typus)
        value = entry.value if isinstance(entry, Entry) else entry
        return converter(value=value, **typus.named_arguments)

    def encode(self, value: Any, entry_typus: EntryTypus = None) -> str:
        if entry_typus is not None:
            converter = self.get_converter(entry_typus)
        else:
            converter = self.get_converter(type(value))
        return converter(value, mode="encode")

    def convert(self, entry: "Entry", typus: Union[type, EntryTypus]) -> Any:
        if not isinstance(typus, EntryTypus):
            return self._convert_by_type(entry=entry, typus=typus)

        return self._convert_by_entry_typus(entry=entry, typus=typus)

    def __call__(self, entry: "Entry", typus: Union[type, EntryTypus]) -> Any:

        return self.convert(entry=entry, typus=typus)

    def get_converter(self, typus: Union[type, EntryTypus]) -> Callable:
        if not isinstance(typus, EntryTypus):
            typus = EntryTypus(base_typus=typus)

        converter = self.get(typus.base_typus)
        return partial(converter, **typus.named_arguments)
# region[Main_Exec]


if __name__ == '__main__':
    pass

# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from enum import Enum, auto
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.enums import BaseGidEnum

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


class SpecialTypus(Enum):
    AUTO = auto()
    RAW = auto()
    DELAYED = auto()


class SpecAttribute(BaseGidEnum):
    CONVERTER = "converter"
    DESCRIPTION = "description"
    SHORT_DESCRIPTION = "short_description"


# region[Main_Exec]


if __name__ == '__main__':
    pass

# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Any, Union
from pathlib import Path

# * Third Party Imports --------------------------------------------------------------------------------->
import attr

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_config.enums import SpecialTypus

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


@attr.s(auto_attribs=True, auto_detect=True, slots=True)
class EntryTypus:
    original_value: str = attr.ib(on_setattr=attr.setters.NO_OP, default=None)
    base_typus: type = attr.ib(on_setattr=attr.setters.NO_OP, default=SpecialTypus.DELAYED)
    named_arguments: dict[str, Any] = attr.ib(default=None, converter=attr.converters.default_if_none(factory=dict))

    def __hash__(self) -> int:
        return hash(self.base_typus)

    def __getitem__(self, key: Union[str, int]) -> Any:
        if isinstance(key, int):
            return self.other_arguments[key]

        return self.named_arguments[key]

    def get(self, key: Union[str, int], default: Any = None) -> Any:
        try:
            return self[key]
        except KeyError:
            return default

    def pop(self, key: Union[str, int], default: Any = None) -> Any:
        if isinstance(key, int):
            return self.other_arguments.pop(key, default)
        return self.named_arguments.pop(key, default)

    def __repr__(self) -> str:
        if self.base_typus is SpecialTypus.DELAYED:
            return f"{self.base_typus.name}(original_value={self.original_value!r})"
        text = f"{self.base_typus}"
        if self.named_arguments:
            text += "(" + ', '.join(f"{key}={value}" for key, value in self.named_arguments.items()) + ')'
        return text

    @property
    def typus(self) -> type:
        sub_typus = [item for item in self.named_arguments.values() if isinstance(item, EntryTypus)]
        if sub_typus:
            sub_typus = sub_typus[0]
            return self.base_typus[sub_typus.typus]
        return self.base_typus

    @classmethod
    @property
    def special_name_conversion_table(cls) -> str:
        return {'str': 'string',
                'int': 'integer',
                'bool': 'boolean'}

    def convert_for_json(self) -> str:

        if self.original_value is not None:
            return self.original_value
        _out = self.base_typus.__name__
        _out = self.special_conversion_map.get(_out, _out)
        args = ''
        if self.named_arguments:
            args = ', '.join(value for value in self.named_arguments.values())

        if args:
            _out += f"({args})"
        return _out


# region[Main_Exec]
if __name__ == '__main__':
    pass

# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
from enum import unique
from typing import Union, AnyStr, Literal
from hashlib import md5, sha256, blake2b, blake2s, sha3_512
from pathlib import Path
from threading import Event, RLock

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.enums import BaseGidEnum
from gidapptools.gid_signal.interface import get_signal
from gidapptools.general_helper.conversion import human2bytes

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()


# endregion[Constants]


READ_TYPE = Union[Literal["r"], Literal["rb"]]
WRITE_TYPE = Union[Literal["w"], Literal["wb"], Literal['a'], Literal['ab']]
ON_ERRORS_TYPE = Union[Literal['replace'], Literal['ignore'], AnyStr]
HASH_FUNC_TYPE = Union[blake2b, md5, sha256, sha3_512, blake2s]


class FileMixin(os.PathLike):
    _encoding = 'utf-8'
    _on_errors: ON_ERRORS_TYPE = 'ignore'

    hash_func: HASH_FUNC_TYPE = md5
    file_hash_size_threshold: int = human2bytes("100 mb")
    path_locks: dict[Path, RLock] = {}
    path_events: dict[Path, Event] = {}

    @unique
    class ChangeParameter(BaseGidEnum):
        SIZE = "size"
        FILE_HASH = "file_hash"
        CHANGED_TIME = "changed_time"
        ALWAYS = "always"
        NEVER = "never"
        ALL = "all"

    def __init__(self, file_path: Path, changed_parameter: str = None, **kwargs) -> None:
        self.file_path = Path(file_path)
        self.name = self.file_path.name.casefold()
        self.changed_parameter = self.ChangeParameter.SIZE if changed_parameter is None else self.ChangeParameter(changed_parameter)
        self.read_mode: READ_TYPE = 'r'
        self.write_mode: WRITE_TYPE = 'w'
        self.last_size: int = None
        self.last_file_hash: str = None
        self.last_changed_time: int = None
        self.changed_signal = get_signal(key=self.file_path)
        super().__init__(**kwargs)

    def set_changed_parameter(self, changed_parameter: Union["ChangeParameter", str]) -> None:
        if isinstance(changed_parameter, self.ChangeParameter):
            self.changed_parameter = changed_parameter
        else:
            self.changed_parameter = self.ChangeParameter(changed_parameter)

    @property
    def lock(self) -> RLock:
        if self.file_path not in self.path_locks:
            self.path_locks[self.file_path] = RLock()
        return self.path_locks[self.file_path]

    @property
    def disable_read_event(self) -> Event:
        if self.file_path not in self.path_events:
            self.path_events[self.file_path] = Event()
        return self.path_events[self.file_path]

    @property
    def size(self) -> int:
        size = self.file_path.stat().st_size
        return size

    @property
    def file_hash(self) -> str:

        with self.file_path.open('rb') as f:
            if self.size <= self.file_hash_size_threshold:

                return self.hash_func(f.read()).hexdigest()
            _file_hash = self.hash_func()
            for chunk in f:
                _file_hash.update(chunk)
            return _file_hash.hexdigest()

    @property
    def changed_time(self) -> int:
        return self.file_path.stat().st_mtime

    @property
    def has_changed(self) -> bool:
        if self.disable_read_event.is_set() is True:
            return False

        def on_size() -> bool:
            return self.last_size is None or self.last_size != self.size

        def on_file_hash() -> bool:
            return self.last_file_hash is None or self.last_file_hash != self.file_hash

        def on_time() -> bool:
            return self.last_changed_time is None or self.last_changed_time != self.changed_time

        def on_all() -> bool:
            return any([on_size(), on_file_hash(), on_time()])

        def on_always() -> bool:
            return True

        def on_never() -> bool:
            return False
        checks = {self.ChangeParameter.SIZE: on_size,
                  self.ChangeParameter.FILE_HASH: on_file_hash,
                  self.ChangeParameter.CHANGED_TIME: on_time,
                  self.ChangeParameter.ALL: on_all,
                  self.ChangeParameter.ALWAYS: on_always,
                  self.ChangeParameter.NEVER: on_never}
        with self.lock:
            result = checks[self.changed_parameter]()
            if result is True:
                self.changed_signal.emit(self)
        return result

    def _update_changed_data(self) -> None:

        def _update_size():
            self.last_size = self.size

        def _update_file_hash():
            self.last_file_hash = self.file_hash

        def _update_changed_time():
            self.last_changed_time = self.changed_time

        def _update_all():
            _update_size()
            _update_file_hash()
            _update_changed_time()
        update_table = {self.ChangeParameter.NEVER: lambda: ...,
                        self.ChangeParameter.ALWAYS: lambda: ...,
                        self.ChangeParameter.SIZE: _update_size,
                        self.ChangeParameter.FILE_HASH: _update_file_hash,
                        self.ChangeParameter.CHANGED_TIME: _update_changed_time,
                        self.ChangeParameter.ALL: _update_all}
        update_table[self.changed_parameter]()

    @property
    def _read_kwargs(self) -> dict[str, str]:
        kwargs = {"mode": self.read_mode}
        if 'b' not in self.read_mode:
            kwargs['encoding'] = self._encoding
            kwargs['errors'] = self._on_errors
        return kwargs

    @property
    def _write_kwargs(self) -> dict[str, str]:
        kwargs = {"mode": self.write_mode}
        if 'b' not in self.write_mode:
            kwargs['encoding'] = self._encoding
            kwargs['errors'] = self._on_errors
        return kwargs

    def read(self):
        with self.lock:
            self._update_changed_data()
            # pylint: disable=unspecified-encoding
            with self.file_path.open(**self._read_kwargs) as f:
                return f.read()

    def write(self, data) -> None:
        with self.lock:
            # pylint: disable=unspecified-encoding
            with self.file_path.open(**self._write_kwargs) as f:
                f.write(data)

    # def __getattr__(self, name: str) -> Any:
    #     if hasattr(self.file_path, name):
    #         return getattr(self.file_path, name)
    #     return super().__getattr__(name)

    def __fspath__(self) -> str:
        return str(self.file_path)

    def __str__(self) -> str:
        return self.__fspath__()
# region[Main_Exec]


if __name__ == '__main__':
    pass
# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import sys
import queue
import atexit
import logging
from typing import Union, Iterable
from pathlib import Path
from logging.handlers import QueueHandler, QueueListener

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_logger.enums import LoggingLevel
from gidapptools.gid_logger.handler import GidBaseRotatingFileHandler
from gidapptools.gid_logger.formatter import GidLoggingFormatter, get_all_func_names, get_all_module_names

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


class GidLogger(logging.Logger):
    ...


def _modify_logger_name(name: str) -> str:
    if name == "__main__":
        return 'main'
    name = 'main.' + '.'.join(name.split('.')[1:])
    return name


def get_logger(name: str) -> Union[logging.Logger, GidLogger]:
    name = _modify_logger_name(name)
    return logging.getLogger(name)


def get_main_logger(name: str, path: Path, log_level: LoggingLevel = LoggingLevel.DEBUG, formatter: Union[logging.Formatter, GidLoggingFormatter] = None, extra_logger: Iterable[str] = tuple()) -> Union[logging.Logger, GidLogger]:
    os.environ["MAX_FUNC_NAME_LEN"] = str(min([max(len(i) for i in get_all_func_names(path, True)), 20]))
    os.environ["MAX_MODULE_NAME_LEN"] = str(min([max(len(i) for i in get_all_module_names(path)), 20]))

    handler = logging.StreamHandler(stream=sys.stdout)

    que = queue.Queue(-1)
    que_handler = QueueHandler(que)
    listener = QueueListener(que, handler)
    formatter = GidLoggingFormatter() if formatter is None else formatter
    handler.setFormatter(formatter)
    _log = get_logger(name)
    for logger in [_log] + [logging.getLogger(l) for l in extra_logger]:
        logger.addHandler(que_handler)

        logger.setLevel(log_level)
    _log.addHandler(que_handler)
    _log.setLevel(log_level)
    listener.start()
    atexit.register(listener.stop)
    return _log


def get_main_logger_with_file_logging(name: str,
                                      log_file_base_name: str,
                                      path: Path,
                                      log_level: LoggingLevel = LoggingLevel.DEBUG,
                                      formatter: Union[logging.Formatter, GidLoggingFormatter] = None,
                                      log_folder: Path = None,
                                      extra_logger: Iterable[str] = tuple()) -> Union[logging.Logger, GidLogger]:
    if os.getenv('IS_DEV', "false") != "false":
        log_folder = path.parent.joinpath('logs')

    os.environ["MAX_FUNC_NAME_LEN"] = str(min([max(len(i) for i in get_all_func_names(path, True)), 20]))
    os.environ["MAX_MODULE_NAME_LEN"] = str(min([max(len(i) for i in get_all_module_names(path)), 20]))
    que = queue.Queue(-1)
    que_handler = QueueHandler(que)

    handler = logging.StreamHandler(stream=sys.stdout)
    file_handler = GidBaseRotatingFileHandler(base_name=log_file_base_name, log_folder=log_folder)

    formatter = GidLoggingFormatter() if formatter is None else formatter
    handler.setFormatter(formatter)
    file_handler.setFormatter(formatter)
    listener = QueueListener(que, handler, file_handler)
    _log = get_logger(name)
    for logger in [_log] + [logging.getLogger(l) for l in extra_logger]:
        logger.addHandler(que_handler)

        logger.setLevel(log_level)
    listener.start()
    atexit.register(listener.stop)
    return _log

# region[Main_Exec]


if __name__ == '__main__':
    pass

# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import inspect
from abc import abstractmethod
from enum import Enum
from typing import Any, Mapping, TypeVar, Callable, Optional
from pathlib import Path
from datetime import datetime, timezone
from importlib.metadata import metadata

# * Third Party Imports --------------------------------------------------------------------------------->
import psutil
from yarl import URL
from appdirs import AppDirs

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.types import PATH_TYPE
from gidapptools.utility.enums import NamedMetaPath
from gidapptools.general_helper.date_time import DatetimeFmt

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


def abstract_class_property(func):
    return property(classmethod(abstractmethod(func)))


def utc_now():
    return datetime.now(tz=timezone.utc)


def handle_path(path: Optional[PATH_TYPE]):
    if path is None:
        return path
    return Path(path).resolve()


def memory_in_use():
    memory = psutil.virtual_memory()
    return memory.total - memory.available


def meta_data_from_path(in_path: Path) -> dict[str, Any]:
    _init_module = inspect.getmodule(None, in_path)
    _metadata = metadata(_init_module.__package__)
    return {k.casefold(): v for k, v in _metadata.items()}


TCallable = TypeVar("TCallable", bound=Callable)


def mark_appdir_path(func: TCallable) -> TCallable:
    func._appdir_path_type = NamedMetaPath(func.__name__)
    return func


class PathLibAppDirs(AppDirs):
    mark_path = mark_appdir_path

    def __init__(self,
                 appname: str,
                 appauthor: str = None,
                 version: str = None,
                 roaming: bool = True,
                 multipath: bool = False) -> None:
        super().__init__(appname=appname, appauthor=appauthor, version=version, roaming=roaming, multipath=multipath)

    @mark_appdir_path
    def user_data_dir(self) -> Path:
        return Path(super().user_data_dir)

    @mark_appdir_path
    def user_log_dir(self) -> Path:
        return Path(super().user_log_dir)

    @mark_appdir_path
    def user_cache_dir(self) -> Path:
        return Path(super().user_cache_dir)

    @mark_appdir_path
    def user_config_dir(self) -> Path:
        return Path(super().user_config_dir)

    @mark_appdir_path
    def user_state_dir(self) -> Path:
        return Path(super().user_state_dir)

    @mark_appdir_path
    def site_data_dir(self) -> Path:
        return Path(super().site_data_dir)

    @mark_appdir_path
    def site_config_dir(self) -> Path:
        return Path(super().site_config_dir)

    def as_path_dict(self) -> dict[NamedMetaPath, Optional[Path]]:
        path_dict = {named_path_item: None for named_path_item in NamedMetaPath.__members__.values()}
        for meth_name, meth_object in inspect.getmembers(self, inspect.ismethod):
            if hasattr(meth_object, '_appdir_path_type'):
                path_dict[meth_object._appdir_path_type] = meth_object()
        return path_dict


def make_pretty(inst) -> dict:

    # pylint: disable=too-many-return-statements
    def _make_pretty(obj):

        def _handle_iterable(in_obj):
            return [_make_pretty(item) for item in in_obj]

        if isinstance(obj, Enum):
            return str(obj)

        if isinstance(obj, Path):
            return obj.as_posix()
        if isinstance(obj, URL):
            return str(obj)
        if isinstance(obj, datetime):
            return DatetimeFmt.STANDARD.strf(obj)

        if isinstance(obj, Mapping):
            pretty_dict = {}
            for key, value in obj.items():
                pretty_dict[_make_pretty(key)] = getattr(inst, f"pretty_{key}", _make_pretty(value))
            return pretty_dict

        if isinstance(obj, list):
            return _handle_iterable(obj)

        if isinstance(obj, tuple):
            return tuple(_handle_iterable(obj))
        if isinstance(obj, set):
            return set(_handle_iterable(obj))
        if isinstance(obj, frozenset):
            return frozenset(_handle_iterable(obj))
        return obj
    return _make_pretty(vars(inst))


def get_qualname_or_name(in_object: Any) -> str:
    try:
        return in_object.__qualname__
    except AttributeError:
        return in_object.__name__


# region[Main_Exec]
if __name__ == '__main__':
    a = PathLibAppDirs(appauthor='giddi', appname='check_appdir')
    print(a.as_path_dict())
# endregion[Main_Exec]



from pathlib import Path
import os
IMAGES_DIR = Path(__file__).parent.absolute()


class StoredImage:

    def __init__(self, path: Path):
        self.path = path.resolve()
        self.name = self.path.stem
        self.image_format = self.path.suffix.removeprefix('.')
        self._bytes: bytes = None

    @property
    def bytes(self) -> bytes:
        if self._bytes is None:
            self._bytes = self.path.read_bytes()
        return self._bytes


_IMAGE_CACHE: dict[str, StoredImage] = {}

PLACEHOLDER_IMAGE = StoredImage(IMAGES_DIR.joinpath("placeholder.png"))


def get_image(name: str) -> StoredImage:
    if name in _IMAGE_CACHE:
        return _IMAGE_CACHE[name]
    for dirname, folderlist, filelist in os.walk(IMAGES_DIR):
        for file in filelist:
            if file.casefold() == name:
                path = Path(dirname, file)
                image = StoredImage(path)
                _IMAGE_CACHE[image.name] = image
                return image



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path

# * Third Party Imports --------------------------------------------------------------------------------->
import click
from rich.table import Table
from rich.markup import escape
from rich.console import Console as RichConsole

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.output_helper.rich_helper import dict_to_rich_tree

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()
CONSOLE = RichConsole(soft_wrap=True, record=True)
# endregion[Constants]


@click.group(name="appmeta")
def appmeta_cli():
    ...


@appmeta_cli.command(help="Lists all available Plugins")
def plugins():
    from gidapptools.meta_data.interface import app_meta
    table = Table(title="[b u light_steel_blue3]AVAILABLE APPMETA-PLUGINS[/b u light_steel_blue3]")
    table.add_column("Plugin Item", style="i chartreuse2 on grey15", header_style="b white on grey37", justify="center", no_wrap=True)
    table.add_column('Module', style="b light_slate_blue on grey15", header_style="b white on grey37", justify="center", no_wrap=True)
    table.add_column('File', style="u dark_khaki on grey15", header_style="b white on grey37", justify="center", no_wrap=True)

    for data in app_meta.plugin_data:
        table.add_row(data.get("product_name"), data.get("module"), f":open_file_folder: [link file://{data.get('file')}]{escape(data.get('file'))}")
    CONSOLE.print("")
    CONSOLE.print(table)


@appmeta_cli.command()
def base_settings():
    from gidapptools.meta_data.interface import app_meta

    CONSOLE.print(dict_to_rich_tree('Base Settings', app_meta.default_base_configuration))


# region[Main_Exec]
if __name__ == '__main__':
    pass
# endregion[Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Any, Union, Callable, Iterable, Optional
from pathlib import Path
from functools import partial
from threading import RLock

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import EntryMissingError, SectionMissingError, ValueValidationError, MissingTypusOrSpecError
from gidapptools.gid_config.enums import SpecialTypus
from gidapptools.general_helper.enums import MiscEnum
from gidapptools.gid_config.parser.tokens import Section
from gidapptools.gid_config.parser.ini_parser import BaseIniParser
from gidapptools.gid_config.parser.config_data import ConfigFile
from gidapptools.gid_config.conversion.spec_data import SpecFile, SpecVisitor, SpecAttribute
from gidapptools.gid_config.conversion.conversion_table import ConfigValueConversionTable
from gidapptools.gid_config.conversion.entry_typus_item import EntryTypus

# endregion[Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion[Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion[Constants]


class SectionAccessor:

    def __init__(self, config: "GidIniConfig", section_name: str) -> None:
        self.config = config
        self.section_name = section_name

    def get(self,
            entry_key: str,
            typus: Union[type, EntryTypus] = SpecialTypus.AUTO,
            fallback_entry: Iterable[str] = None,
            default: Any = MiscEnum.NOTHING) -> Any:
        return self.config.get(section_name=self.section_name, entry_key=entry_key, typus=typus, fallback_entry=fallback_entry, default=default)

    def set(self,
            entry_key: str,
            entry_value: Any,
            create_missing_section: bool = False,
            spec_typus: str = None) -> None:
        return self.config.set(section_name=self.section_name, entry_key=entry_key, entry_value=entry_value, create_missing_section=create_missing_section, spec_typus=spec_typus)


class GidIniConfig:
    access_locks_storage: dict[tuple, RLock] = {}
    default_spec_visitor: type[SpecVisitor] = SpecVisitor
    default_parser: type[BaseIniParser] = BaseIniParser
    default_converter: type[ConfigValueConversionTable] = ConfigValueConversionTable

    def __init__(self,
                 config_file: Path,
                 config_file_auto_write: bool = True,
                 spec_file: Optional[SpecFile] = None,
                 converter: ConfigValueConversionTable = None,
                 empty_is_missing: bool = True,
                 spec_visitor: SpecVisitor = None,
                 parser: BaseIniParser = None,
                 file_changed_parameter: str = 'size') -> None:

        self.parser = self.default_parser() if parser is None else parser
        self.spec_visitor = self.default_spec_visitor() if spec_visitor is None else spec_visitor
        self.config = ConfigFile(file_path=config_file, parser=self.parser, changed_parameter=file_changed_parameter, auto_write=config_file_auto_write)
        self.spec = SpecFile(file_path=spec_file, visitor=self.spec_visitor, changed_parameter=file_changed_parameter) if spec_file is not None else None
        self.converter = self.default_converter() if converter is None else converter
        self.empty_is_missing = empty_is_missing

    @property
    def access_lock(self) -> RLock:
        spec_path = self.spec.file_path if self.spec is not None else None
        key = (self.config.file_path, spec_path)
        if key not in self.access_locks_storage:
            lock = RLock()
            self.access_locks_storage[key] = lock
        else:
            lock = self.access_locks_storage[key]
        return lock

    def get_description(self, section_name: str, entry_key: str) -> str:
        return self.spec.get(key_path=[section_name, entry_key, SpecAttribute.DESCRIPTION.value], default="")

    def reload(self) -> None:
        with self.access_lock:
            self.config.reload()
            self.spec.reload()

    def get_section(self, section_name: str) -> dict[str, Any]:
        with self.access_lock:
            section = self.config.get_section(section_name=section_name)
            return {entry.key: self.get(section_name=section_name, entry_key=entry.key) for entry in section.entries.values()}

    def get_section_accessor(self, section_name: str) -> Callable[[str, Optional[Union[type, EntryTypus]], Optional[Iterable[str]], Optional[Any]], Any]:
        with self.access_lock:
            return partial(self.get, section_name=section_name)

    def get_entry_accessor(self, section_name: str,
                           entry_key: str,
                           typus: Union[type, EntryTypus] = SpecialTypus.AUTO,
                           fallback_entry: Iterable[str] = None,
                           default: Any = MiscEnum.NOTHING):
        return partial(self.get, section_name=section_name, entry_key=entry_key, typus=typus, fallback_entry=fallback_entry, default=default)

    def get(self,
            section_name: str,
            entry_key: str,
            typus: Union[type, EntryTypus] = SpecialTypus.AUTO,
            fallback_entry: Iterable[str] = None,
            default: Any = MiscEnum.NOTHING) -> Any:
        with self.access_lock:
            try:
                entry = self.config.get_entry(section_name=section_name, entry_key=entry_key)

            except (EntryMissingError, SectionMissingError):
                if fallback_entry is not None:
                    return self.get(fallback_entry[0], fallback_entry[1], default=default)
                if default is not MiscEnum.NOTHING:
                    return default
                raise

            if not entry.value:
                if self.empty_is_missing is True:
                    if fallback_entry is not None:
                        return self.get(fallback_entry[0], fallback_entry[1], default=default)
                    if default is not MiscEnum.NOTHING:
                        return default
                spec_default = self.spec._get_entry_default(section_name=section_name, entry_key=entry_key)
                if spec_default not in {MiscEnum.NOTHING, None}:
                    entry.value = spec_default
                else:
                    return None

            if typus is SpecialTypus.AUTO:
                if self.spec is None:
                    raise MissingTypusOrSpecError("You have to provide a typus if no spec file has been set in the __init__.")
                typus = self.spec.get_entry_typus(section_name=section_name, entry_key=entry_key)
            try:
                return self.converter(entry=entry, typus=typus)
            except ValueValidationError:
                if default is not MiscEnum.NOTHING:
                    return default
                raise

    def set(self,
            section_name: str,
            entry_key: str,
            entry_value: Any,
            create_missing_section: bool = False,
            spec_typus: str = None) -> None:
        with self.access_lock:
            self.config.disable_read_event.set()
            try:
                entry_typus = self.spec.get_entry_typus(section_name=section_name, entry_key=entry_key)
                self.config.set_value(section_name=section_name, entry_key=entry_key, entry_value=self.converter.encode(entry_value, entry_typus=entry_typus), create_missing_section=create_missing_section)
                if spec_typus is not None:
                    self.spec.set_typus_value(section_name=section_name, entry_key=entry_key, typus_value=spec_typus)
            finally:
                self.config.disable_read_event.clear()

    def clear_section(self, section_name: str, missing_ok: bool = True) -> None:
        with self.access_lock:
            self.config.clear_entries(section_name=section_name, missing_ok=missing_ok)

    def remove_section(self, section_name: str, missing_ok: bool = True) -> None:
        with self.access_lock:
            self.config.remove_section(section_name=section_name, missing_ok=missing_ok)

    def add_section(self, section_name: str, existing_ok: bool = True) -> None:
        with self.access_lock:
            self.config.add_section(section=Section(section_name), existing_ok=existing_ok)

    def add_spec_value_handler(self, target_name: str, handler: Callable[[Any, list[Any]], EntryTypus]) -> None:
        with self.access_lock:
            self.spec_visitor.add_handler(target_name=target_name, handler=handler)

    def add_converter_function(self, typus: Any, converter_function: Callable) -> None:
        with self.access_lock:
            self.converter[typus] = converter_function

    def as_dict(self, raw: bool = False, with_typus: bool = False) -> dict[str, dict[str, Any]]:
        with self.access_lock:
            raw_dict = self.config.as_raw_dict()
            if raw is True:
                return raw_dict
            _out = {}
            for section_name, values in raw_dict.items():
                _out[section_name] = {}
                for entry_name in values:
                    if with_typus is True:
                        _out[section_name][entry_name] = (self.get(section_name, entry_name), self.get_entry_typus(section_name, entry_name))
                    else:
                        _out[section_name][entry_name] = self.get(section_name, entry_name)

            return _out

    def get_entry_typus(self, section_name: str, entry_key: str):
        return self.spec.get_entry_typus(section_name=section_name, entry_key=entry_key)


# region[Main_Exec]
if __name__ == '__main__':
    pass
    # x = GidIniConfig(config_file=Path(r"D:\Dropbox\hobby\Modding\Programs\Github\My_Repos\GidAppTools\tests\gid_config_tests\example_config_1.ini"),
    #                  spec_file=Path(r"D:\Dropbox\hobby\Modding\Programs\Github\My_Repos\GidAppTools\tests\gid_config_tests\example_spec_1.json"))
    # x.reload()
# endregion[Main_Exec]
